from .input_types import *
from .exceptions import ObjectNotExistException, MutationValidationException
from .mutation_by_filter import mutation_on_uuids_from_filter, mutation_on_uuids_from_filter_business_model



import decimal
import json
import logging
import re
import sys
import uuid
from copy import copy
from datetime import datetime as py_datetime
from functools import reduce

import graphene_django_optimizer as gql_optimizer
from core.services import create_or_update_interactive_user, create_or_update_core_user, create_or_update_officer, \
    create_or_update_claim_admin, change_user_password
from core.tasks import openimis_mutation_async
from django import dispatch
from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError, PermissionDenied
from django.core.serializers.json import DjangoJSONEncoder
from django.db import transaction
from django.db.models import Q
from django.db.models.expressions import RawSQL
from django.http import HttpRequest
from django.utils import translation
from graphene.utils.str_converters import to_snake_case
from graphene_django.filter import DjangoFilterConnectionField
import graphql_jwt
from typing import Optional

from .apps import CoreConfig
from .gql_queries import *
from .models import ModuleConfiguration, FieldControl, MutationLog, Language, RoleMutation, UserMutation

MAX_SMALLINT = 32767
MIN_SMALLINT = -32768

core = sys.modules["core"]

logger = logging.getLogger(__name__)


class SmallInt(graphene.Int):
    """
    This represents a small Integer, with values ranging from -32768 to +32767
    """

    @staticmethod
    def coerce_int(value):
        res = super().coerce_int(value)
        if MIN_SMALLINT <= res <= MAX_SMALLINT:
            return res
        else:
            return None

    serialize = coerce_int
    parse_value = coerce_int

    @staticmethod
    def parse_literal(ast):
        result = graphene.Int.parse_literal(ast)
        if result is not None and MIN_SMALLINT <= result <= MAX_SMALLINT:
            return result
        else:
            return None


MAX_TINYINT = 255
MIN_TINYINT = 0


class TinyInt(graphene.Int):
    """
    This represents a tiny Integer (8 bit), with values ranging from 0 to 255
    """

    @staticmethod
    def coerce_int(value):
        res = super().coerce_int(value)
        if MIN_TINYINT <= res <= MAX_TINYINT:
            return res
        else:
            return None

    serialize = coerce_int
    parse_value = coerce_int

    @staticmethod
    def parse_literal(ast):
        result = graphene.Int.parse_literal(ast)
        if result is not None and MIN_TINYINT <= result <= MAX_TINYINT:
            return result
        else:
            return None


class OpenIMISJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        if isinstance(o, HttpRequest):
            if o.user:
                return f"HTTP_user: {o.user.id}"
            else:
                return None
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super().default(o)


_mutation_signal_params = ["user", "mutation_module",
                           "mutation_class", "mutation_log_id", "data"]
signal_mutation = dispatch.Signal(providing_args=_mutation_signal_params)
signal_mutation_module_validate = {}
signal_mutation_module_before_mutating = {}
signal_mutation_module_after_mutating = {}

for module in sys.modules:
    signal_mutation_module_validate[module] = dispatch.Signal(providing_args=_mutation_signal_params)
    signal_mutation_module_before_mutating[module] = dispatch.Signal(providing_args=_mutation_signal_params)
    signal_mutation_module_after_mutating[module] = \
        dispatch.Signal(providing_args=_mutation_signal_params + ["error_messages"])


class OpenIMISMutation(graphene.relay.ClientIDMutation):
    """
    This class is the generic Mutation for openIMIS. It will save the mutation content into the MutationLog,
    submit it to validation, perform the potentially asynchronous mutation itself and update the MutationLog status.
    """

    class Meta:
        abstract = True

    internal_id = graphene.Field(graphene.String)

    class Input:
        client_mutation_label = graphene.String(max_length=255, required=False)
        client_mutation_details = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data) -> Optional[str]:
        """
        This method has to be overridden in the subclasses to implement the actual mutation.
        The response should contain a boolean for success and an error message that will be saved into the DB
        :param user: contains the logged user or None
        :param data: all parameters passed to the mutation
        :return: error_message if None is returned, the response is considered to be a success
        """
        pass

    @classmethod
    def mutate_and_get_payload(cls, root, info, **data):
        mutation_log = MutationLog.objects.create(
            json_content=json.dumps(data, cls=OpenIMISJSONEncoder),
            user_id=info.context.user.id if info.context.user else None,
            client_mutation_id=data.get("client_mutation_id"),
            client_mutation_label=data.get("client_mutation_label"),
            client_mutation_details=json.dumps(data.get(
                "client_mutation_details"), cls=OpenIMISJSONEncoder) if data.get("client_mutation_details") else None
        )
        logger.debug("OpenIMISMutation: saved as %s, label: %s", mutation_log.id, mutation_log.client_mutation_label)
        if info and info.context and info.context.user and info.context.user.language:
            lang = info.context.user.language
            if isinstance(lang, Language):
                translation.activate(lang.code)
            else:
                translation.activate(lang)

        try:
            logger.debug("[OpenIMISMutation %s] Sending signals", mutation_log.id)
            results = signal_mutation.send(
                sender=cls, mutation_log_id=mutation_log.id, data=data, user=info.context.user,
                mutation_module=cls._mutation_module, mutation_class=cls._mutation_class)
            results.extend(signal_mutation_module_validate[cls._mutation_module].send(
                sender=cls, mutation_log_id=mutation_log.id, data=data, user=info.context.user,
                mutation_module=cls._mutation_module, mutation_class=cls._mutation_class
            ))
            errors = [err for r in results for err in r[1]]
            logger.debug("[OpenIMISMutation %s] signals sent, got errors back: ", mutation_log.id, len(errors))
            if errors:
                mutation_log.mark_as_failed(json.dumps(errors))
                return cls(internal_id=mutation_log.id)

            signal_mutation_module_before_mutating[cls._mutation_module].send(
                sender=cls, mutation_log_id=mutation_log.id, data=data, user=info.context.user,
                mutation_module=cls._mutation_module, mutation_class=cls._mutation_class
            )
            logger.debug("[OpenIMISMutation %s] before mutate signal sent", mutation_log.id)
            if core.async_mutations:
                logger.debug("[OpenIMISMutation %s] Sending async mutation", mutation_log.id)
                openimis_mutation_async.delay(
                    mutation_log.id, cls._mutation_module, cls._mutation_class)
            else:
                logger.debug("[OpenIMISMutation %s] mutating...", mutation_log.id)
                try:
                    error_messages = cls.async_mutate(
                        info.context.user if info.context and info.context.user else None,
                        **data)
                    if not error_messages:
                        logger.debug("[OpenIMISMutation %s] marked as successful", mutation_log.id)
                        mutation_log.mark_as_successful()
                    else:
                        errors_json = json.dumps(error_messages)
                        logger.error("[OpenIMISMutation %s] marked as failed: %s", mutation_log.id, errors_json)
                        mutation_log.mark_as_failed(errors_json)
                except BaseException as exc:
                    error_messages = exc
                    logger.error("async_mutate threw an exception. It should have gotten this far.", exc_info=exc)
                    # Record the failure of the mutation but don't include details for security reasons
                    mutation_log.mark_as_failed(f"The mutation threw a {type(exc)}, check logs for details")
                logger.debug("[OpenIMISMutation %s] send post mutation signal", mutation_log.id)
                signal_mutation_module_after_mutating[cls._mutation_module].send(
                    sender=cls, mutation_log_id=mutation_log.id, data=data, user=info.context.user,
                    mutation_module=cls._mutation_module, mutation_class=cls._mutation_class,
                    error_messages=error_messages
                )
        except Exception as exc:
            logger.error(f"Exception while processing mutation id {mutation_log.id}", exc_info=exc)
            mutation_log.mark_as_failed(exc)

        return cls(internal_id=mutation_log.id)


class FieldControlGQLType(DjangoObjectType):
    class Meta:
        model = FieldControl


class ModuleConfigurationGQLType(DjangoObjectType):
    class Meta:
        model = ModuleConfiguration


class OrderedDjangoFilterConnectionField(DjangoFilterConnectionField):
    """
    Adapted from https://github.com/graphql-python/graphene/issues/251
    And then adapted by Alok Ramteke on my (Eric Darchis)' stackoverflow question:
    https://stackoverflow.com/questions/57478464/django-graphene-relay-order-by-orderingfilter/61543302
    Substituting:
    `mutation_logs = DjangoFilterConnectionField(MutationLogGQLType)`
    with:
    ```
    mutation_logs = OrderedDjangoFilterConnectionField(MutationLogGQLType,
        orderBy=graphene.List(of_type=graphene.String))
    ```
    """

    @classmethod
    def _filter_order_by(cls, order_by: str) -> str:
        if order_by:
            return re.sub("[^\\w\\-,+]", "", order_by)
        else:
            return order_by

    @classmethod
    def orderBy(cls, qs, args):
        order = args.get('orderBy', None)
        if order:
            if type(order) is str:
                if order == "?":
                    snake_order = RawSQL("NEWID()", params=[])
                else:
                    # due to https://github.com/advisories/GHSA-xpfp-f569-q3p2 we are aggressively filtering the orderBy
                    snake_order = to_snake_case(cls._filter_order_by(order))
            else:
                snake_order = [
                    to_snake_case(cls._filter_order_by(o)) if o != "?" else RawSQL(
                        "NEWID()", params=[])
                    for o in order
                ]
            qs = qs.order_by(*snake_order)
        return qs

    @classmethod
    def resolve_queryset(
            cls, connection, iterable, info, args, filtering_args, filterset_class
    ):
        qs = super(DjangoFilterConnectionField, cls).resolve_queryset(
            connection, iterable, info, args
        )
        filter_kwargs = {k: v for k, v in args.items() if k in filtering_args}
        qs = filterset_class(data=filter_kwargs, queryset=qs, request=info.context).qs

        return OrderedDjangoFilterConnectionField.orderBy(qs, args)


class MutationLogGQLType(DjangoObjectType):
    """
    This represents a requested mutation and its status.
    The "user" search filter is only available for super-users. Otherwise, the user is automatically set to the
    currently logged user.
    """

    class Meta:
        model = MutationLog
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "client_mutation_id": ["exact"],
            "client_mutation_label": ["exact"],
            "request_date_time": ["exact", "gte", "lte"],
            "status": ["exact", "gte"],
            "user": ["exact"],
        }
        connection_class = ExtendedConnection

    status = graphene.Field(graphene.Int,
                            description=", ".join([f"{pair[0]}: {pair[1]}" for pair in MutationLog.STATUS_CHOICES]))

    @classmethod
    def get_queryset(cls, queryset, info):
        if info.context.user.is_anonymous:
            return queryset.none()
        elif info.context.user.is_superuser:
            return queryset
        else:
            queryset = queryset.filter(user=info.context.user)
        return queryset


UT_INTERACTIVE = "INTERACTIVE"
UT_TECHNICAL = "TECHNICAL"
UT_OFFICER = "OFFICER"
UT_CLAIM_ADMIN = "CLAIM_ADMIN"

UserTypeEnum = graphene.Enum("UserTypes", [
            (UT_INTERACTIVE, UT_INTERACTIVE),
            (UT_OFFICER, UT_OFFICER),
            (UT_TECHNICAL, UT_TECHNICAL),
            (UT_CLAIM_ADMIN, UT_CLAIM_ADMIN)
        ])


class Query(graphene.ObjectType):
    module_configurations = graphene.List(
        ModuleConfigurationGQLType,
        validity=graphene.String(),
        layer=graphene.String())

    mutation_logs = OrderedDjangoFilterConnectionField(
        MutationLogGQLType, orderBy=graphene.List(of_type=graphene.String))

    role = OrderedDjangoFilterConnectionField(
        RoleGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        is_system=graphene.Boolean(),
        show_history=graphene.Boolean(),
        client_mutation_id=graphene.String(),
        str=graphene.String(description="Text search on any field")
    )

    role_right = OrderedDjangoFilterConnectionField(
        RoleRightGQLType, orderBy=graphene.List(of_type=graphene.String), validity=graphene.Date()
    )

    interactiveUsers = OrderedDjangoFilterConnectionField(
        InteractiveUserGQLType, orderBy=graphene.List(of_type=graphene.String), validity=graphene.Date(),
        show_history=graphene.Boolean(),
        client_mutation_id=graphene.String(),
    )

    users = OrderedDjangoFilterConnectionField(
        UserGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        validity=graphene.Date(),
        client_mutation_id=graphene.String(),
        last_name=graphene.String(description="partial match, case insensitive"),
        other_names=graphene.String(description="partial match, case insensitive"),
        phone=graphene.String(description="exact match on phone number"),
        email=graphene.String(description="exact match on email address"),
        role_id=graphene.Int(),
        roles=graphene.List(of_type=graphene.Int),
        health_facility_id=graphene.Int(description="Base health facility ID (not UUID!)"),
        region_id=graphene.Int(),
        district_id=graphene.Int(),
        municipality_id=graphene.Int(),
        village_id=graphene.Int(),
        birth_date_from=graphene.Date(),
        birth_date_to=graphene.Date(),
        user_types=graphene.List(of_type=UserTypeEnum),
        language=graphene.String(),
        str=graphene.String(description="text search that will check username, last name, other names and email"),
        description="This interface provides access to the various types of users in openIMIS. The main resource"
                    "is limited to a username and refers either to a TechnicalUser or InteractiveUser. Only the latter"
                    "is exposed in GraphQL. There are also optional links to ClaimAdministrator and Officer depending"
                    "on the setup. BEWARE, fetching these links is costly as there is no direct database link between"
                    "these entities and there are retrieved one by one. Do not fetch them for large lists if you can"
                    "avoid it. The showHistory is acting on the InteractiveUser, avoid mixing with Officer or "
                    "ClaimAdmin."
    )

    user = OrderedDjangoFilterConnectionField(
        UserGQLType, order_by=graphene.List(of_type=graphene.String)
    )

    enrolment_officers = OrderedDjangoFilterConnectionField(
        OfficerGQLType,
        str=graphene.String(
            description="text search that will check username, last name, other names and email"
        ),
    )

    modules_permissions = graphene.Field(
        ModulePermissionsListGQLType,
    )

    languages = graphene.List(LanguageGQLType)

    def resolve_enrolment_officers(self, info, **kwargs):
        from .models import Officer

        if not info.context.user.has_perms(
            CoreConfig.gql_query_enrolment_officers_perms
        ):
            raise PermissionError("Unauthorized")

        search = kwargs.get("str")

        if search is not None:
            return gql_optimizer.query(
                Officer.objects.filter(
                    Q(code__icontains=search)
                    | Q(last_name__icontains=search)
                    | Q(other_names__icontains=search)
                ),
                info,
            )

    def resolve_interactive_users(self, info, **kwargs):
        if not info.context.user.has_perms(CoreConfig.gql_query_users_perms):
            raise PermissionError("Unauthorized")
        filters = []
        query = InteractiveUser.objects

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        show_history = kwargs.get('show_history', False)
        if not show_history and not kwargs.get('uuid', None):
            filters += filter_validity(**kwargs)

        return gql_optimizer.query(query.filter(*filters), info)

    def resolve_users(self, info, email=None, last_name=None, other_names=None, phone=None,
                      role_id=None, roles=None, health_facility_id=None, region_id=None, district_id=None,
                      municipality_id=None, birth_date_from=None, birth_date_to=None, user_types=None, language=None,
                      village_id=None, **kwargs):
        if not info.context.user.has_perms(CoreConfig.gql_query_users_perms):
            raise PermissionError("Unauthorized")

        user_filters = []
        user_query = User.objects.exclude(t_user__isnull=False)
        text_search = kwargs.get("str")  # Poorly chosen name, avoid of shadowing "str"
        if text_search:
            user_filters.append(Q(username__icontains=text_search) |
                                Q(i_user__last_name__icontains=text_search) |
                                Q(officer__last_name__icontains=text_search) |
                                Q(claim_admin__last_name__icontains=text_search) |
                                Q(i_user__other_names__icontains=text_search) |
                                Q(officer__other_names__icontains=text_search) |
                                Q(claim_admin__other_names__icontains=text_search) |
                                Q(i_user__email=text_search) |
                                Q(officer__email=text_search) |
                                Q(claim_admin__email_id=text_search)
                                )

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            user_filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        if email:
            user_filters.append(Q(i_user__email=email) |
                                Q(officer__email=email) |
                                Q(claim_admin__email_id=email))
        if phone:
            user_filters.append(Q(i_user__phone=phone) |
                                Q(officer__phone=phone) |
                                Q(claim_admin__phone=phone))
        if last_name:
            user_filters.append(Q(i_user__last_name__icontains=last_name) |
                                Q(officer__last_name__icontains=last_name) |
                                Q(claim_admin__last_name__icontains=last_name))
        if other_names:
            user_filters.append(Q(i_user__other_names__icontains=other_names) |
                                Q(officer__other_names__icontains=other_names) |
                                Q(claim_admin__other_names__icontains=other_names))
        if language:
            user_filters.append(Q(i_user__language=language))
            # Language is not applicable to Office/ClaimAdmin
        if health_facility_id:
            user_filters.append(Q(i_user__health_facility_id=health_facility_id) |
                                Q(officer__location_id=health_facility_id) |
                                Q(claim_admin__health_facility_id=health_facility_id))
        if birth_date_from:
            user_filters.append(Q(officer__dob__gte=birth_date_from) |
                                Q(officer__veo_dob__gte=birth_date_from) |
                                Q(claim_admin__dob__gte=birth_date_from))
        if birth_date_to:
            user_filters.append(Q(officer__dob__lte=birth_date_to) |
                                Q(officer__veo_dob__lte=birth_date_to) |
                                Q(claim_admin__dob__lte=birth_date_to))
        if role_id:
            user_filters.append(Q(i_user__user_roles__role_id=role_id) &
                                Q(i_user__user_roles__validity_to__isnull=True))
        if roles:
            user_filters.append(Q(i_user__user_roles__role_id__in=roles) &
                                Q(i_user__user_roles__validity_to__isnull=True))
        if region_id:
            user_filters.append(Q(i_user__userdistrict__location__parent_id=region_id))
        if district_id:
            user_filters.append(Q(i_user__userdistrict__location_id=district_id))
        if municipality_id:
            user_filters.append(Q(officer__officer_villages__location__parent_id=municipality_id))
        if village_id:
            user_filters.append(Q(officer__officer_villages__location_id=village_id))

        if user_types:
            ut_conditions = {
                UT_INTERACTIVE: Q(i_user__isnull=False),
                UT_OFFICER: Q(officer__isnull=False),
                UT_TECHNICAL: Q(t_user__isnull=False),
                UT_CLAIM_ADMIN: Q(claim_admin__isnull=False),
            }
            user_filters.append(reduce(lambda a, b: a | b, [ut_conditions[x] for x in user_types]))

        # Do NOT use the query optimizer here ! It would make the t_user, officer etc as deferred fields if they are not
        # explicitly requested in the GraphQL response. However, this prevents the dynamic remapping of the User object.
        return user_query.filter(*user_filters)

    def resolve_role(self, info, **kwargs):
        if not info.context.user.has_perms(CoreConfig.gql_query_roles_perms):
            raise PermissionError("Unauthorized")
        filters = []
        query = Role.objects

        text_search = kwargs.get("str")
        if text_search:
            filters.append(Q(name__icontains=text_search))

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        show_history = kwargs.get('show_history', False)
        if not show_history and not kwargs.get('uuid', None):
            filters += filter_validity(**kwargs)

        is_system_role = kwargs.get('is_system', None)
        # check if we can use default filter validity
        if is_system_role is not None:
            if is_system_role:
                query = query.filter(
                    is_system__gte=1
                )
            else:
                query = query.filter(
                    is_system=0
                )
        return gql_optimizer.query(query.filter(*filters), info)

    def resolve_role_right(self, info, **kwargs):
        if not info.context.user.has_perms(CoreConfig.gql_query_roles_perms):
            raise PermissionError("Unauthorized")
        return gql_optimizer.query(RoleRight.objects.all(), info)

    def resolve_modules_permissions(self, info, **kwargs):
        if not info.context.user.has_perms(CoreConfig.gql_query_roles_perms):
            raise PermissionError("Unauthorized")
        excluded_app = [
            "health_check.cache", "health_check", "health_check.db",
            "test_without_migrations", "test_without_migrations",
            "rules", "graphene_django", "rest_framework", "rest_framework_rules",
            "health_check.storage", "channels", "graphql_jwt.refresh_token.apps.RefreshTokenConfig"
        ]
        all_apps = [app for app in settings.INSTALLED_APPS if not app.startswith("django") and app not in excluded_app]
        config = []
        for app in all_apps:
            apps = __import__(f"{app}.apps")
            if hasattr(apps.apps, 'DEFAULT_CFG'):
                config_dict = ModuleConfiguration.get_or_default(f"{app}", apps.apps.DEFAULT_CFG)
                permission = []
                for key, value in config_dict.items():
                    if key.endswith("_perms"):
                        if isinstance(value, list):
                            for val in value:
                                permission.append(PermissionOpenImisGQLType(
                                    perms_name=key,
                                    perms_value=val,
                                ))
                config.append(ModulePermissionGQLType(
                    module_name=app,
                    permissions=permission,
                ))
        return ModulePermissionsListGQLType(list(config))

    def resolve_module_configurations(self, info, **kwargs):
        validity = kwargs.get('validity')
        # configuration is loaded before even the core module
        # the datetime is ALWAYS a Gregorian one
        # (whatever datetime is used in business modules)
        if validity is None:
            validity = py_datetime.now()
        else:
            d = re.split('\D', validity)
            validity = py_datetime(*[int('0' + x) for x in d][:6])
        # is_exposed indicates wherever a configuration
        # is safe to be accessible from api
        # DON'T EXPOSE (backend) configurations that contain credentials,...
        crits = (
            Q(is_disabled_until=None) | Q(is_disabled_until__lt=validity),
            Q(is_exposed=True)
        )
        layer = kwargs.get('layer')
        if layer is not None:
            crits = (*crits, Q(layer=layer))
        return ModuleConfiguration.objects.prefetch_related('controls').filter(*crits)

    def resolve_languages(self, info, **kwargs):
        return Language.objects.order_by('sort_order').all()


class RoleBase:
    id = graphene.Int(required=False, read_only=True)
    uuid = graphene.String(required=False)
    name = graphene.String(required=True, max_length=50)
    alt_language = graphene.String(required=False, max_length=50)
    is_system = graphene.Boolean(required=True)
    is_blocked = graphene.Boolean(required=True)
    # field to save all chosen rights to the role
    rights_id = graphene.List(graphene.Int, required=False)


def update_or_create_role(data, user):
    client_mutation_id = data.get("client_mutation_id", None)
    # client_mutation_label = data.get("client_mutation_label", None)

    if "client_mutation_id" in data:
        data.pop('client_mutation_id')
    if "client_mutation_label" in data:
        data.pop('client_mutation_label')
    role_uuid = data.pop('uuid') if 'uuid' in data else None
    rights_id = data.pop('rights_id') if "rights_id" in data else None
    if role_uuid:
        role = Role.objects.get(uuid=role_uuid)
        role.save_history()
        [setattr(role, k, v) for k, v in data.items()]
        role.save()
        if rights_id is not None:
            # reset all role rights assigned to the chosen role
            from core import datetime
            now = datetime.datetime.now()
            role_rights_currently_assigned = RoleRight.objects.filter(role_id=role.id)
            role_rights_currently_assigned.update(validity_to=now)
            role_rights_currently_assigned = role_rights_currently_assigned.values_list('right_id', flat=True)
            for right_id in rights_id:
                if right_id not in role_rights_currently_assigned:
                    # create role right because it is a new role right
                    RoleRight.objects.create(
                        role_id=role.id,
                        right_id=right_id,
                        audit_user_id=role.audit_user_id,
                        validity_from=now,
                    )
                else:
                    # set date valid to - None
                    role_right = RoleRight.objects.get(Q(role_id=role.id, right_id=right_id))
                    role_right.validity_to = None
                    role_right.save()
    else:
        role = Role.objects.create(**data)
        # create role rights for that role if they were passed to mutation
        if rights_id:
            [RoleRight.objects.create(
                **{
                    "role_id": role.id,
                    "right_id": right_id,
                    "audit_user_id": role.audit_user_id,
                    "validity_from": data['validity_from'],
                }
            ) for right_id in rights_id]
        if client_mutation_id:
            RoleMutation.object_mutated(user, role=role, client_mutation_id=client_mutation_id)
        return role
    return role


def duplicate_role(data, user):
    client_mutation_id = data.get("client_mutation_id", None)
    # client_mutation_label = data.get("client_mutation_label", None)

    if "client_mutation_id" in data:
        data.pop('client_mutation_id')
    if "client_mutation_label" in data:
        data.pop('client_mutation_label')
    role_uuid = data.pop('uuid') if 'uuid' in data else None
    rights_id = data.pop('rights_id') if "rights_id" in data else None
    # get the current Role object to be duplicated
    role = Role.objects.get(uuid=role_uuid)
    # copy Role to be dupliacated
    from core import datetime
    now = datetime.datetime.now()
    duplicated_role = copy(role)
    duplicated_role.id = None
    duplicated_role.uuid = uuid.uuid4()
    duplicated_role.validity_from = now
    [setattr(duplicated_role, k, v) for k, v in data.items()]
    duplicated_role.save()
    if rights_id:
        # reset all role rights assigned to the chosen role
        role_rights_currently_assigned = RoleRight.objects.filter(role_id=role.id)
        role_rights_currently_assigned = role_rights_currently_assigned.values_list('right_id', flat=True)
        for right_id in rights_id:
            validity_from = now
            if right_id in role_rights_currently_assigned:
                # role right exist - we can assign validity_from from old entity
                validity_from = role.validity_from
            # create role right for duplicate role
            RoleRight.objects.create(
                **{
                    "role_id": duplicated_role.id,
                    "right_id": right_id,
                    "audit_user_id": duplicated_role.audit_user_id,
                    "validity_from": validity_from,
                }
            )
    else:
        role_rights_currently_assigned = RoleRight.objects.filter(role_id=role.id)
        [RoleRight.objects.create(
            **{
                "role_id": duplicated_role.id,
                "right_id": role_right.right_id,
                "audit_user_id": duplicated_role.audit_user_id,
                "validity_from": now,
            }
        ) for role_right in role_rights_currently_assigned]

    if client_mutation_id:
        RoleMutation.object_mutated(user, role=duplicated_role, client_mutation_id=client_mutation_id)

    return duplicated_role


class CreateRoleMutation(OpenIMISMutation):
    """
    Create a new role, with its chosen role right
    """
    _mutation_module = "core"
    _mutation_class = "CreateRoleMutation"

    class Input(RoleBase, OpenIMISMutation.Input):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError("mutation.authentication_required")
            if not user.has_perms(CoreConfig.gql_mutation_create_roles_perms):
                raise PermissionDenied("unauthorized")
            from core.utils import TimeUtils
            data['validity_from'] = TimeUtils.now()
            data['audit_user_id'] = user.id_for_audit
            update_or_create_role(data, user)
            return None
        except Exception as exc:
            return [
                {
                    'message': "core.mutation.failed_to_create_role",
                    'detail': str(exc)
                }]


class UpdateRoleMutation(OpenIMISMutation):
    """
    Update a chosen role, with its chosen role right
    """
    _mutation_module = "core"
    _mutation_class = "UpdateRoleMutation"

    class Input(RoleBase, OpenIMISMutation.Input):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError("mutation.authentication_required")
            if not user.has_perms(CoreConfig.gql_mutation_update_roles_perms):
                raise PermissionDenied("unauthorized")
            if 'uuid' not in data:
                raise ValidationError("There is no uuid in updateMutation input!")
            data['audit_user_id'] = user.id_for_audit
            update_or_create_role(data, user)
            return None
        except Exception as exc:
            return [
                {
                    'message': "core.mutation.failed_to_update_role",
                    'detail': str(exc)
                }]


def set_role_deleted(role):
    try:
        role.delete_history()
        return []
    except Exception as exc:
        return {
            'title': role.uuid,
            'list': [{
                'message': "role.mutation.failed_to_change_status_of_role" % {'role': str(role)},
                'detail': role.uuid}]
        }


class DeleteRoleMutation(OpenIMISMutation):
    """
        Delete a chosen role
    """
    _mutation_module = "core"
    _mutation_class = "DeleteRoleMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(CoreConfig.gql_mutation_delete_roles_perms):
            raise PermissionDenied("unauthorized")
        errors = []
        for role_uuid in data["uuids"]:
            role = Role.objects \
                .filter(uuid=role_uuid) \
                .first()
            if role is None:
                errors.append({
                    'title': role,
                    'list': [{'message':
                        "role.validation.id_does_not_exist" % {'id': role_uuid}}]
                })
                continue
            errors += set_role_deleted(role)
        if len(errors) == 1:
            errors = errors[0]['list']
        return errors


class DuplicateRoleMutation(OpenIMISMutation):
    """
        Duplicate a chosen role
    """
    _mutation_module = "core"
    _mutation_class = "DuplicateRoleMutation"

    class Input(OpenIMISMutation.Input):
        uuid = graphene.String(required=True)
        name = graphene.String(required=True, max_length=50)
        alt_language = graphene.String(required=False, max_length=50)
        is_system = graphene.Boolean(required=True)
        is_blocked = graphene.Boolean(required=True)
        # field to save all chosen rights to the role
        rights_id = graphene.List(graphene.Int, required=False)

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError("mutation.authentication_required")
            if not user.has_perms(CoreConfig.gql_mutation_duplicate_roles_perms):
                raise PermissionDenied("unauthorized")
            data['audit_user_id'] = user.id_for_audit
            duplicate_role(data, user)
            return None
        except Exception as exc:
            return [
                {
                    'message': "core.mutation.failed_to_duplicate_role",
                    'detail': str(exc)
                }]


class UserBase:
    uuid = graphene.String(
        required=False,
        read_only=True,
        description="UUID of the core User, one can leave this blank and specify the username instead",
    )
    user_id = graphene.String(required=False)
    other_names = graphene.String(required=True, max_length=50)
    last_name = graphene.String(required=True, max_length=50)
    username = graphene.String(required=True, max_length=8)
    phone_number = graphene.String(required=False)
    email = graphene.String(required=False)
    password = graphene.String(required=False)
    health_facility_id = graphene.Int(required=False)
    districts = graphene.List(graphene.Int, required=False)
    language = graphene.String(required=True, description="Language code for the user")
    # Interactive User only
    roles = graphene.List(
        graphene.Int,
        required=False,
        description="List of role_ids, required for interactive users",
    )

    # Enrolment Officer / Feedback / Claim Admin specific
    birth_date = graphene.Date(required=False)
    address = graphene.String(required=False)  # multi-line
    works_to = graphene.DateTime(required=False)
    substitution_officer_id = graphene.Int(required=False)
    # TODO VEO_code, last_name, other names, dob, phone
    phone_communication = graphene.Boolean(required=False)
    location_id = graphene.Int(
        required=False, description="Location for the Enrolment Officer"
    )
    village_ids = graphene.List(graphene.Int, required=False)

    user_types = graphene.List(UserTypeEnum, required=True)


class CreateUserMutation(OpenIMISMutation):
    """
    Create a new user, the "core" one but also Interactive, Technical, Officer or Admin
    """
    _mutation_module = "core"
    _mutation_class = "CreateUserMutation"

    class Input(UserBase, OpenIMISMutation.Input):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError("mutation.authentication_required")
            if not user.has_perms(CoreConfig.gql_mutation_create_roles_perms):
                raise PermissionDenied("unauthorized")
            from core.utils import TimeUtils
            data['validity_from'] = TimeUtils.now()
            data['audit_user_id'] = user.id_for_audit
            update_or_create_user(data, user)
            return None
        except Exception as exc:
            return [
                {
                    'message': "core.mutation.failed_to_create_user",
                    'detail': str(exc)
                }]


class UpdateUserMutation(OpenIMISMutation):
    """
    Update an existing User and sub-user types
    """
    _mutation_module = "core"
    _mutation_class = "UpdateUserMutation"

    class Input(UserBase, OpenIMISMutation.Input):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError("mutation.authentication_required")
            if not user.has_perms(CoreConfig.gql_mutation_update_roles_perms):
                raise PermissionDenied("unauthorized")
            from core.utils import TimeUtils
            data['validity_from'] = TimeUtils.now()
            data['audit_user_id'] = user.id_for_audit
            update_or_create_user(data, user)
            return None
        except Exception as exc:
            return [
                {
                    'message': "core.mutation.failed_to_update_user",
                    'detail': str(exc)
                }]


class DeleteUserMutation(OpenIMISMutation):
    """
    Delete a chosen user
    """
    _mutation_module = "core"
    _mutation_class = "DeleteUserMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        if not user.has_perms(CoreConfig.gql_mutation_delete_users_perms):
            raise PermissionDenied("unauthorized")
        errors = []
        for user_uuid in data["uuids"]:
            user = User.objects \
                .filter(id=user_uuid) \
                .first()
            if user is None:
                errors.append({
                    'title': user,
                    'list': [{'message':
                        "user.validation.id_does_not_exist" % {'id': user_uuid}}]
                })
                continue
            errors += set_user_deleted(user)
        if len(errors) == 1:
            errors = errors[0]['list']
        return errors


@transaction.atomic
def update_or_create_user(data, user):
    client_mutation_id = data.get("client_mutation_id", None)
    # client_mutation_label = data.get("client_mutation_label", None)

    if "client_mutation_id" in data:
        data.pop('client_mutation_id')
    if "client_mutation_label" in data:
        data.pop('client_mutation_label')
    user_uuid = data.pop('uuid') if 'uuid' in data else None

    if UT_INTERACTIVE in data["user_types"]:
        i_user, i_user_created = create_or_update_interactive_user(
            user_uuid, data, user.id_for_audit, len(data["user_types"]) > 1)
    else:
        i_user, i_user_created = None, False
    if UT_OFFICER in data["user_types"]:
        officer, officer_created = create_or_update_officer(
            user_uuid, data, user.id_for_audit, UT_INTERACTIVE in data["user_types"])
    else:
        officer, officer_created = None, False
    if UT_CLAIM_ADMIN in data["user_types"]:
        claim_admin, claim_admin_created = create_or_update_claim_admin(
            user_uuid, data, user.id_for_audit, UT_INTERACTIVE in data["user_types"])
    else:
        claim_admin, claim_admin_created = None, False
    core_user, core_user_created = create_or_update_core_user(
        user_uuid=user_uuid, username=data["username"], i_user=i_user, officer=officer, claim_admin=claim_admin)

    if client_mutation_id:
        UserMutation.object_mutated(user, core_user=core_user, client_mutation_id=client_mutation_id)
    return core_user


def set_user_deleted(user):
    try:
        if user.i_user:
            user.i_user.delete_history()
        if user.t_user:
            user.t_user.delete_history()
        if user.officer:
            user.officer.delete_history()
        if user.claim_admin:
            user.claim_admin.delete_history()
        user.delete()  # TODO: we might consider disabling Users instead of deleting entirely.
        return []
    except Exception as exc:
        logger.info("role.mutation.failed_to_change_status_of_user" % {'user': str(user)})
        return {
            "title": user.id,
            "list": [{
                "message": "role.mutation.failed_to_change_status_of_user" % {'user': str(user)},
                "detail": user.id}]
        }


class ChangePasswordMutation(OpenIMISMutation):
    """
    Change a user's password. Either the user can update his own by providing the old password, or an administrator
    (actually someone with the rights to update users) can force it for anyone without providing the old password.
    """
    _mutation_module = "core"
    _mutation_class = "ChangePasswordMutation"

    class Input(OpenIMISMutation.Input):
        username = graphene.String(
            required=False,
            description="By default, this operation works on the logged user,"
                        "only administrators can run it on any user")
        old_password = graphene.String(
            required=False,
            description="Mandatory to change the current user password, administrators can leave this blank"
        )
        new_password = graphene.String(
            required=True,
            description="New password to set"
        )

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            if type(user) is AnonymousUser or not user.id:
                raise ValidationError("mutation.authentication_required")

            change_user_password(
                user,
                username_to_update=data.get("username"),
                old_password=data.get("old_password"),
                new_password=data.get("new_password"),
            )
            return None
        except Exception as exc:
            return [
                {
                    'message': "core.mutation.failed_to_change_user_password",
                    'detail': str(exc)
                }]


class Mutation(graphene.ObjectType):
    create_role = CreateRoleMutation.Field()
    update_role = UpdateRoleMutation.Field()
    delete_role = DeleteRoleMutation.Field()
    duplicate_role = DuplicateRoleMutation.Field()

    create_user = CreateUserMutation.Field()
    update_user = UpdateUserMutation.Field()
    delete_user = DeleteUserMutation.Field()
    change_password = ChangePasswordMutation.Field()

    token_auth = graphql_jwt.mutations.ObtainJSONWebToken.Field()
    verify_token = graphql_jwt.mutations.Verify.Field()
    refresh_token = graphql_jwt.mutations.Refresh.Field()
    revoke_token = graphql_jwt.mutations.Revoke.Field()



def on_role_mutation(sender, **kwargs):
    uuid = kwargs['data'].get('uuid', None)
    if not uuid:
        return []

    # For duplicate log is created in the duplicate_role function, mutation log added here would reference original role
    if "Role" in str(sender._mutation_class) and sender._mutation_class != 'DuplicateRoleMutation':
        impacted = Role.objects.get(uuid=uuid)
        RoleMutation.objects.create(
            role=impacted, mutation_id=kwargs['mutation_log_id']
        )

    return []


def on_user_mutation(sender, **kwargs):
    uuid = kwargs['data'].get('uuid', None)
    if not uuid:
        return []

    # For duplicate log is created in the duplicate_role function, mutation log added here would reference original role
    if "User" in str(sender._mutation_class):
        impacted = User.objects.get(id=uuid)
        UserMutation.objects.create(
            core_user=impacted, mutation_id=kwargs["mutation_log_id"]
        )

    return []


def bind_signals():
    signal_mutation_module_validate["core"].connect(on_role_mutation)
    signal_mutation_module_validate["core"].connect(on_user_mutation)



from django.dispatch import Signal as BaseSignal
from django.dispatch.dispatcher import _make_id
import functools

from typing import Callable, Dict

from core.service_signals import RegisteredServiceSignal, ServiceSignalBindType


class Signal(BaseSignal):

    def connect(self, receiver, sender=None, weak=True, dispatch_uid=None, priority=50):
        if dispatch_uid is None:
            dispatch_uid = _make_id(receiver)

        inner_uid = '{0}{1}'.format(priority, dispatch_uid)
        super(Signal, self).connect(receiver, sender=sender, weak=weak, dispatch_uid=inner_uid)
        self.receivers.sort()


REGISTERED_SERVICE_SIGNALS: Dict[str, RegisteredServiceSignal] = {}


def __register_service_signal(signal_name, providing_args):
    signal = REGISTERED_SERVICE_SIGNALS.get(signal_name)
    if signal_name in REGISTERED_SERVICE_SIGNALS:
        if signal.is_signal_registered():
            raise AttributeError(F'Signal name {signal_name} clashes with already existing '
                                 F'signal name {REGISTERED_SERVICE_SIGNALS[signal_name]} and can\'t be registered.')
        else:
            signal.register_signal(providing_args)
    else:
        REGISTERED_SERVICE_SIGNALS[signal_name] = RegisteredServiceSignal(providing_args)


def __raise_unregistered_signal_exception(signal_name):
    raise ValueError(f"Signal for name {signal_name} not found. Was it registered?"
                     f"Note: Classes/functions wrapped in register_service_signal have to be imported in "
                     f"signals.py, as Django doesn't fetch all classes definitions at initializations. Therefore, "
                     f"signal is only registered after import and class definition.")


def register_service_signal(signal_name: str):
    """
    Decorator for registering signals for service. Two signals will be created: one before service and one after.
    Signals are stored in core.signals.REGISTERED_SERVICE_SIGNALS as RegisteredServiceSignal.
    See RegisteredServiceSignal class definition for more information regarding when signals are registered
    and connected. Signal receivers output is stored in RegisteredServiceSignal.signal_results. Results are
    overridden after very call.

    Example use:
        @register_service_signal('create_policy', PolicyServiceClass):
        def create_policy(self, *args, **kwargs)
            ...

        Will create new RegisteredServiceSignal instance available from  REGISTERED_SERVICE_SIGNALS['create_policy'].
        Signal name has to be unique.

    """
    __register_service_signal(signal_name, ['cls_', 'data', 'context'])

    def decorator(func):
        @functools.wraps(func)
        def wrapper_do_twice(*args, **kwargs):
            registered_signal = REGISTERED_SERVICE_SIGNALS.get(signal_name, None)
            if registered_signal is None:
                __raise_unregistered_signal_exception(signal_name)

            context = kwargs.pop('context', None)
            cls_, func_args = args[0], args[1:]  # First element of args is self/cls. Moved to separate variable.
            signal_call_args = {'cls_': cls_, 'data': [func_args, kwargs], 'context': context}

            registered_signal.send_signal_before(sender=cls_, **signal_call_args)

            out = func(*args, **kwargs)

            signal_call_args['result'] = out
            registered_signal.send_signal_after(sender=cls_, **signal_call_args)
            return out
        return wrapper_do_twice

    return decorator


def bind_service_signal(signal_name: str, func: Callable,
                        bind_type: ServiceSignalBindType = ServiceSignalBindType.BEFORE_AND_AFTER):
    """
    By default binding is done after modules are loaded, with same similar approach as for graphql.
    Main OpenIMIS backend is crawling through modules searching for bind_service_signals function
    in module_name.signals path. In most cases function connected to signal will be loaded before signals are
    registered. They're queued and connected after signal definition is provided.

    Example:
        Content of openimis-be-invoice_py/invoice/signals.py:
        from core.signals import bind_service_signal
        def bind_service_signals():
            bind_service_signal('create_signal', connected_signal_function)

    @param signal_name:
    @param func:
    @param bind_type:
    @return:
    """
    if signal_name not in REGISTERED_SERVICE_SIGNALS.keys():
        __register_service_signal(signal_name, None)

    signal = REGISTERED_SERVICE_SIGNALS[signal_name]
    signal.connect_signal(func, bind_type)



from django.db.models import Lookup

from jsonfallback.fields import FallbackJSONField


@FallbackJSONField.register_lookup
class JsonContains(Lookup):
    lookup_name = 'jsoncontains'

    sql_server_json_key_prefix = '$'
    sql_server_nested_json_separator = '.'
    BASE_SQL = '(JSON_VALUE(cast({} as nvarchar(max)), %s)=%s)'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)

        rhs_params[0].prepare(connection)
        params = self._build_sql_params(lhs, rhs_params[0].adapted)
        sql_statement = ' AND '.join([self.BASE_SQL.format(lhs) for _ in range(0, len(rhs_params[0].adapted))])
        return sql_statement, params

    def _prepare_dict_value(self, json_ext):
        def flatten(dictionary, separator, prefix):
            return {
                prefix + separator + key if prefix else key: value
                for nested_key, nested_value in dictionary.items()
                for key, value in flatten(nested_value, separator, nested_key).items()
            } if isinstance(dictionary, dict) else {prefix: str(dictionary)}

        base_prefix = self.sql_server_json_key_prefix
        base_separator = self.sql_server_nested_json_separator
        return flatten(json_ext, base_separator, base_prefix)

    def _build_sql_params(self, entity, json_conditions, ):
        adjusted_conditions = self._prepare_dict_value(json_conditions)
        conditions = []
        for json_key, expected_value in adjusted_conditions.items():
            # conditions.append(entity)
            conditions.append(json_key)
            conditions.append(expected_value)
        return conditions



import logging

from apscheduler.schedulers.background import BackgroundScheduler
# from apscheduler.executors.pool import ProcessPoolExecutor, ThreadPoolExecutor
from core import get_scheduler_method_ref
from django_apscheduler.jobstores import register_events  # , register_job

from django.conf import settings

logger = logging.getLogger(__name__)

# Create scheduler to run in a thread inside the application process
scheduler = BackgroundScheduler(settings.SCHEDULER_CONFIG)


def schedule_tasks(task_scheduler):
    """
    Does the actual scheduling and is shared between the start() below and the management command for standalone
    execution
    :param scheduler: scheduler to which we'll add the tasks
    """
    if settings.SCHEDULER_JOBS:
        for job in settings.SCHEDULER_JOBS:
            logger.debug("Scheduling job %s", job["method"])
            method = get_scheduler_method_ref(job["method"])
            task_scheduler.add_job(*([method] + job.get("args", [])), **(job.get("kwargs", {})))

    if settings.SCHEDULER_CUSTOM:
        for job in settings.SCHEDULER_CUSTOM:
            logger.debug("Calling custom scheduler %s", job["method"])
            method = get_scheduler_method_ref(job["method"])
            method(*([task_scheduler] + job.get("args", [])), **(job.get("kwargs", {})))


def start():
    if settings.DEBUG:
        # Hook into the apscheduler logger
        logging.basicConfig()
        logging.getLogger('apscheduler').setLevel(logging.DEBUG)

    schedule_tasks(scheduler)

    # Add the scheduled jobs to the Django admin interface
    register_events(scheduler)

    scheduler.start()



import sys
import os
import json
import logging
import uuid
from copy import copy

import core
from django.apps import apps
from django.db import models
from django.db.models import Q, DO_NOTHING, F
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied, ValidationError
from django.contrib.auth.models import AbstractBaseUser, BaseUserManager, PermissionsMixin, Group
from django.utils.crypto import salted_hmac
from cached_property import cached_property
from graphql import ResolveInfo
from .fields import DateTimeField
from datetime import datetime as py_datetime
from .utils import filter_validity, filter_is_deleted
from jsonfallback.fields import FallbackJSONField
from simple_history.models import HistoricalRecords
from dirtyfields import DirtyFieldsMixin

logger = logging.getLogger(__name__)


class UUIDModel(models.Model):
    """
    Abstract entity, parent of all (new) openIMIS entities.
    Enforces the UUID identifier.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True

    def __str__(self):
        return "[%s]" % (self.id,)


class BaseVersionedModel(models.Model):
    validity_from = DateTimeField(db_column='ValidityFrom', default=py_datetime.now)
    validity_to = DateTimeField(db_column='ValidityTo', blank=True, null=True)

    def save_history(self, **kwargs):
        if not self.id:  # only copy if the data is being updated
            return None
        histo = copy(self)
        histo.id = None
        if hasattr(histo, "uuid"):
            setattr(histo, "uuid", uuid.uuid4())
        from core import datetime
        histo.validity_to = datetime.datetime.now()
        histo.legacy_id = self.id
        histo.save()
        return histo.id

    def delete_history(self, **kwargs):
        self.save_history()
        from core import datetime
        now = datetime.datetime.now()
        self.validity_from = now
        self.validity_to = now
        self.save()

    class Meta:
        abstract = True

    @classmethod
    def filter_queryset(cls, queryset=None):
        if queryset is None:
            queryset = cls.objects.all()
        queryset = queryset.filter(*filter_validity())
        return queryset


class VersionedModel(BaseVersionedModel):
    legacy_id = models.IntegerField(
        db_column='LegacyID', blank=True, null=True)

    class Meta:
        abstract = True


class UUIDVersionedModel(BaseVersionedModel):
    legacy_id = models.UUIDField(
        db_column='LegacyID', blank=True, null=True)

    class Meta:
        abstract = True


class ExtendableModel(models.Model):
    json_ext = FallbackJSONField(
        db_column='JsonExt', blank=True, null=True)

    class Meta:
        abstract = True


class ModuleConfiguration(UUIDModel):
    """
    Generic entity to save every modules' configuration (json format)
    """
    module = models.CharField(max_length=20)
    MODULE_LAYERS = [('fe', 'frontend'), ('be', 'backend')]
    layer = models.CharField(
        max_length=2,
        choices=MODULE_LAYERS,
        default='be',
    )
    version = models.CharField(max_length=10)
    config = models.TextField()
    # is_exposed indicates wherever a configuration is safe to be accessible from api
    # DON'T EXPOSE (backend) configurations that contain credentials,...
    is_exposed = models.BooleanField(default=False)
    is_disabled_until = models.DateTimeField(
        default=None,
        blank=True,
        null=True
    )

    @classmethod
    def get_or_default(cls, module, default, layer='be'):
        if bool(os.environ.get('NO_DATABASE', False)):
            logger.info('env NO_DATABASE set to True: ModuleConfiguration not loaded from db!')
            return default

        try:
            now = py_datetime.now()  # can't use core config here...
            db_configuration = cls.objects.get(
                Q(is_disabled_until=None) | Q(is_disabled_until__lt=now),
                layer=layer,
                module=module
            )._cfg
            return {**default, **db_configuration}
        except ModuleConfiguration.DoesNotExist:
            logger.info('No %s configuration, using default!' % module)
            return default
        except Exception:
            logger.error('Failed to load %s configuration, using default!\n%s: %s' % (
                module, sys.exc_info()[0].__name__, sys.exc_info()[1]))
            return default

    @cached_property
    def _cfg(self):
        import collections
        return json.loads(self.config, object_pairs_hook=collections.OrderedDict)

    def __str__(self):
        return "%s [%s]" % (self.module, self.version)

    class Meta:
        db_table = 'core_ModuleConfiguration'


class FieldControl(UUIDModel):
    module = models.ForeignKey(
        ModuleConfiguration, models.DO_NOTHING, related_name='controls')
    field = models.CharField(unique=True, max_length=250)
    # mask: Hidden | Readonly | Mandatory
    usage = models.IntegerField(blank=True, null=True)

    class Meta:
        db_table = 'core_FieldControl'


class Language(models.Model):
    code = models.CharField(db_column='LanguageCode',
                            primary_key=True, max_length=5)
    name = models.CharField(db_column='LanguageName', max_length=50)
    sort_order = models.IntegerField(
        db_column='SortOrder', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'tblLanguages'


class UserManager(BaseUserManager):

    def _create_core_user(self, **fields):
        user = User(**fields)
        user.save()
        return user

    def _create_tech_user(self, username, email, password, **extra_fields):
        tech = TechnicalUser(username=username, email=email, **extra_fields)
        tech.set_password(password)
        tech.save()
        return tech

    def create_user(self, username, password, email=None, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields['is_superuser'] = False
        self._create_tech_user(username, email, password, **extra_fields)

    def create_superuser(self, username, password=None, email=None, **extra_fields):
        extra_fields['is_staff'] = True
        extra_fields['is_superuser'] = True
        self._create_tech_user(username, email, password, **extra_fields)

    def auto_provision_user(self, **kwargs):
        # only auto-provision django user if registered as interactive user
        try:
            i_user = InteractiveUser.objects.get(
                login_name=kwargs['username'],
                *filter_validity())
        except InteractiveUser.DoesNotExist:
            raise PermissionDenied
        user = self._create_core_user(**kwargs)
        user.i_user = i_user
        user.save()
        if core.auto_provisioning_user_group:
            group = Group.objects.get(
                name=core.auto_provisioning_user_group)
            user_group = UserGroup(user=user, group=group)
            user_group.save()
        return user, True

    def get_or_create(self, **kwargs):
        try:
            user = User.objects.get(**kwargs)
            return user, False
        except User.DoesNotExist:
            return self.auto_provision_user(**kwargs)


class TechnicalUser(AbstractBaseUser):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    username = models.CharField(max_length=150, unique=True)
    email = models.EmailField(blank=True, null=True)
    language = 'en'
    is_staff = models.BooleanField(default=False)
    is_superuser = models.BooleanField(default=False)
    validity_from = models.DateTimeField(blank=True, null=True)
    validity_to = models.DateTimeField(blank=True, null=True)

    @property
    def id_for_audit(self):
        return -1

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['password']

    def _bind_User(self):
        save_required = False
        try:
            usr = User.objects.get(t_user=self)
        except ObjectDoesNotExist:
            usr = User()
            usr.t_user = self
            save_required = True
        if usr.username != self.username:
            usr.username = self.username
            save_required = True
        if save_required:
            usr.save()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self._bind_User()

    class Meta:
        managed = True
        db_table = 'core_TechnicalUser'


class Role(VersionedModel):
    id = models.AutoField(db_column='RoleID', primary_key=True)
    uuid = models.CharField(db_column='RoleUUID', max_length=36, default=uuid.uuid4, unique=True)
    name = models.CharField(db_column='RoleName', max_length=50)
    alt_language = models.CharField(
        db_column='AltLanguage', max_length=50, blank=True, null=True)
    is_system = models.IntegerField(db_column='IsSystem')
    is_blocked = models.BooleanField(db_column='IsBlocked')
    audit_user_id = models.IntegerField(
        db_column='AuditUserID', blank=True, null=True)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        managed = False
        db_table = 'tblRole'


class RoleRight(VersionedModel):
    id = models.AutoField(db_column='RoleRightID', primary_key=True)
    role = models.ForeignKey(Role, models.DO_NOTHING,
                             db_column='RoleID', related_name="rights")
    right_id = models.IntegerField(db_column='RightID')
    audit_user_id = models.IntegerField(
        db_column='AuditUserId', blank=True, null=True)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        managed = False
        db_table = 'tblRoleRight'


class InteractiveUser(VersionedModel):
    id = models.AutoField(db_column="UserID", primary_key=True)
    uuid = models.CharField(db_column="UserUUID", max_length=36, default=uuid.uuid4, unique=True)
    language = models.ForeignKey(Language, models.DO_NOTHING, db_column="LanguageID")
    last_name = models.CharField(db_column="LastName", max_length=100)
    other_names = models.CharField(db_column="OtherNames", max_length=100)
    phone = models.CharField(db_column="Phone", max_length=50, blank=True, null=True)
    login_name = models.CharField(db_column="LoginName", max_length=25)
    health_facility_id = models.IntegerField(db_column="HFID", blank=True, null=True)

    audit_user_id = models.IntegerField(db_column="AuditUserID")
    # password = models.BinaryField(blank=True, null=True)
    # dummy_pwd is always blank. It is actually a transient field used in the Legacy to pass the clear text password in
    # a User object from the ASPX to the DAL where it is processed into/against password and private key/salt)
    # dummy_pwd = models.CharField(db_column='DummyPwd', max_length=25, blank=True, null=True)
    email = models.CharField(db_column="EmailId", max_length=200, blank=True, null=True)
    private_key = models.CharField(db_column="PrivateKey", max_length=256, blank=True, null=True,
                                   help_text="The private key is actually a password salt")
    stored_password = models.CharField(db_column="StoredPassword", max_length=256, blank=True, null=True,
                                       help_text="By default a SHA256 of the private key (salt) and password")
    password_validity = models.DateTimeField(db_column="PasswordValidity", blank=True, null=True)
    is_associated = models.BooleanField(db_column="IsAssociated", blank=True, null=True,
                                        help_text="has a claim admin or enrolment officer account")
    role_id = models.IntegerField(db_column="RoleID", null=False)

    @property
    def id_for_audit(self):
        return id

    @property
    def username(self):
        return self.login_name

    def get_username(self):
        return self.login_name

    @property
    def is_staff(self):
        return False

    @property
    def is_superuser(self):
        return False

    @cached_property
    def rights(self):
        return [rr.right_id for rr in RoleRight.filter_queryset().filter(
            role_id__in=[r.role_id for r in UserRole.filter_queryset().filter(
                user_id=self.id)]).distinct()]

    @cached_property
    def rights_str(self):
        return [str(r) for r in self.rights]

    @cached_property
    def get_health_facility(self):
        if self.health_facility_id:
            hf_model = apps.get_model("location", "HealthFacility")
            if hf_model:
                return hf_model.objects.filter(pk=self.health_facility_id).first()

    def set_password(self, raw_password):
        from hashlib import sha256
        from secrets import token_hex
        self.private_key = token_hex(128)
        pwd_hash = sha256()
        pwd_hash.update(f"{raw_password.rstrip()}{self.private_key}".encode())
        self.stored_password = pwd_hash.hexdigest().upper()  # Legacy requires this to be uppercase

    def check_password(self, raw_password):
        from hashlib import sha256
        pwd_hash = sha256()
        pwd_hash.update(f"{raw_password.rstrip()}{self.private_key}".encode())
        pwd_hash = pwd_hash.hexdigest()
        # logger.debug("pwd_hash %s -> %s, stored: %s", f"{raw_password.rstrip()}{self.private_key}", pwd_hash, self.stored_password)
        # hashlib gives a lowercase digest while the legacy gives an uppercase one
        return pwd_hash == self.stored_password.lower()

    @classmethod
    def is_interactive_user(cls, user):
        if isinstance(user, InteractiveUser):
            return user
        elif isinstance(user, User) and user.i_user is not None:
            return user.i_user
        else:
            return None

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        return queryset

    class Meta:
        managed = False
        db_table = 'tblUsers'


class UserRole(VersionedModel):
    id = models.AutoField(db_column='UserRoleID', primary_key=True)
    user = models.ForeignKey(
        InteractiveUser, models.DO_NOTHING, db_column='UserID', related_name="user_roles")
    role = models.ForeignKey(Role, models.DO_NOTHING,
                             db_column='RoleID', related_name="user_roles")
    audit_user_id = models.IntegerField(
        db_column='AudituserID', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'tblUserRole'


class User(UUIDModel, PermissionsMixin):
    username = models.CharField(unique=True, max_length=25)
    t_user = models.ForeignKey(TechnicalUser, on_delete=models.CASCADE, blank=True, null=True)
    i_user = models.ForeignKey(InteractiveUser, on_delete=models.CASCADE, blank=True, null=True)
    officer = models.ForeignKey("Officer", on_delete=models.CASCADE, blank=True, null=True)
    claim_admin = models.ForeignKey("claim.ClaimAdmin", on_delete=models.CASCADE, blank=True, null=True)

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = []

    objects = UserManager()

    def __init__(self, *args, **kwargs):
        super(User, self).__init__(*args, **kwargs)
        try:
            self._u = self.i_user or self.officer or self.claim_admin or self.t_user
        except ValueError:
            deferred = self.get_deferred_fields()
            if any(x in deferred for x in ["i_user", "officer", "claim_admin", "t_user"]):
                logger.error("Some of the user type fields are deferred, this breaks the dynamic loading of "
                             "openIMIS User, if calling from GraphQL, remove the optimizer")
            raise

    @property
    def email(self):
        if self.i_user:
            return self.i_user.email
        if self.officer:
            return self.officer.email
        if self.claim_admin:
            return self.claim_admin.email_id
        return None

    @property
    def id_for_audit(self):
        return self._u.id

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    @property
    def is_staff(self):
        return self._u.is_staff

    @property
    def is_superuser(self):
        return self._u.is_superuser

    @property
    def is_active(self):
        if self._u.validity_from is None and self._u.validity_to is None:
            return True
        from core import datetime
        now = datetime.datetime.now()
        if not self._u.validity_from is None and self._u.validity_from > now:
            return False
        if not self._u.validity_to is None and self._u.validity_to < now:
            return False
        return True

    def has_perm(self, perm, obj=None):
        i_user = self.i_user if obj is None else obj.i_user
        return True if i_user is not None and perm in i_user.rights_str else super(User, self).has_perm(perm, obj)

    def get_session_auth_hash(self):
        key_salt = "core.User.get_session_auth_hash"
        return salted_hmac(key_salt, self.username).hexdigest()

    def get_health_facility(self):
        if self.claim_admin:
            return self.claim_admin.health_facility
        if self.i_user:
            return self.i_user.get_health_facility()
        return None

    def __getattr__(self, name):
        if name == '_u':
            raise ValueError('wrapper has not been initialised')
        if name == '__name__':
            return self.username
        if name == 'get_session_auth_hash':
            return False
        # if not self._u:
        #     return None
        return getattr(self._u, name)

    def __call__(self, *args, **kwargs):
        # if not self._u:
        #     raise ValueError('wrapper has not been initialised')
        return self._u(*args, **kwargs)

    def __str__(self):
        if self.i_user:
            utype = 'i'
        elif self.t_user:
            utype = 't'
        elif self.officer:
            utype = 'o'
        elif self.claim_admin:
            utype = 'c'
        else:
            utype = '?'
        return "(%s) %s [%s]" % (utype, self.username, self.id)

    def save(self, *args, **kwargs):
        if self._u and self._u.id:
            self._u.save()
        super().save(*args, **kwargs)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        managed = True
        db_table = 'core_User'


class UserGroup(models.Model):
    user = models.ForeignKey(User, models.DO_NOTHING)
    group = models.ForeignKey(Group, models.DO_NOTHING)

    class Meta:
        managed = False
        db_table = 'core_User_groups'
        unique_together = (('user', 'group'),)


class Officer(VersionedModel):
    id = models.AutoField(db_column='OfficerID', primary_key=True)
    uuid = models.CharField(db_column='OfficerUUID',
                            max_length=36, default=uuid.uuid4, unique=True)
    code = models.CharField(db_column='Code', max_length=8)
    last_name = models.CharField(db_column='LastName', max_length=100)
    other_names = models.CharField(db_column='OtherNames', max_length=100)
    dob = models.DateField(db_column='DOB', blank=True, null=True)
    phone = models.CharField(db_column='Phone', max_length=50, blank=True, null=True)
    location = models.ForeignKey('location.Location', models.DO_NOTHING, db_column='LocationId', blank=True, null=True)
    substitution_officer = models.ForeignKey('self', models.DO_NOTHING, db_column='OfficerIDSubst', blank=True,
                                             null=True)
    works_to = models.DateTimeField(db_column='WorksTo', blank=True, null=True)
    veo_code = models.CharField(db_column='VEOCode', max_length=8, blank=True, null=True)
    veo_last_name = models.CharField(db_column='VEOLastName', max_length=100, blank=True, null=True)
    veo_other_names = models.CharField(db_column='VEOOtherNames', max_length=100, blank=True, null=True)
    veo_dob = models.DateField(db_column='VEODOB', blank=True, null=True)
    veo_phone = models.CharField(db_column='VEOPhone', max_length=25, blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')
    # rowid = models.TextField(db_column='RowID', blank=True, null=True)   This field type is a guess.
    email = models.CharField(db_column='EmailId', max_length=200, blank=True, null=True)
    phone_communication = models.BooleanField(db_column='PhoneCommunication', blank=True, null=True)
    address = models.CharField(db_column="permanentaddress", max_length=100, blank=True, null=True)
    has_login = models.BooleanField(db_column='HasLogin', blank=True, null=True)

    # user = models.ForeignKey(User, db_column='UserID', blank=True, null=True, on_delete=models.CASCADE)

    def name(self):
        return " ".join(n for n in [self.last_name, self.other_names] if n is not None)

    def __str__(self):
        return "[%s] %s" % (self.code, self.name())

    @property
    def id_for_audit(self):
        return id

    @property
    def username(self):
        return self.code

    def get_username(self):
        return self.code

    @property
    def is_staff(self):
        return False

    @property
    def is_superuser(self):
        return False

    @cached_property
    def rights(self):
        return []

    @cached_property
    def rights_str(self):
        return []

    def set_password(self, raw_password):
        raise NotImplementedError("Shouldn't set a password on an Officer")

    def check_password(self, raw_password):
        return False

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        return queryset

    class Meta:
        managed = False
        db_table = 'tblOfficer'


class MutationLog(UUIDModel):
    """
    Maintains a log of every mutation requested along with its status. It is used to reply
    immediately to the client and have longer processing in the various backend modules.
    The ID of this table will be used for reference.
    """
    RECEIVED = 0
    ERROR = 1
    SUCCESS = 2
    STATUS_CHOICES = (
        (RECEIVED, "Received"),
        (ERROR, "Error"),
        (SUCCESS, "Success"),
    )

    json_content = models.TextField()
    user = models.ForeignKey(User, on_delete=DO_NOTHING, blank=True, null=True)
    request_date_time = models.DateTimeField(auto_now_add=True)
    client_mutation_id = models.CharField(
        max_length=255, blank=True, null=True)
    client_mutation_label = models.CharField(
        max_length=255, blank=True, null=True)
    client_mutation_details = models.TextField(blank=True, null=True)
    status = models.IntegerField(choices=STATUS_CHOICES, default=RECEIVED)
    error = models.TextField(blank=True, null=True)

    class Meta:
        managed = True
        db_table = "core_Mutation_Log"

    def mark_as_successful(self):
        """
        Do not alter the mutation_log and then save it as it might override changes from another process. This
        method will only set the mutation_log as successful if it is in RECEIVED status.
        :return True if the status was updated, False if it was in ERROR or already in SUCCESS status
        """
        affected_rows = MutationLog.objects.filter(id=self.id) \
            .filter(status=MutationLog.RECEIVED).update(status=MutationLog.SUCCESS)
        self.refresh_from_db()
        return affected_rows > 0

    def mark_as_failed(self, error):
        """
        Do not alter the mutation_log and then save it as it might override changes from another process.
        This method will force the status to ERROR and set its error accordingly.
        """
        MutationLog.objects.filter(id=self.id) \
            .update(status=MutationLog.ERROR, error=error)
        self.refresh_from_db()


class ObjectMutation:
    """
    This object is used for link tables between the business objects and the MutationLog like ClaimMutation.
    The object_mutated() method allows the creation of an object to update the xxxMutation easily.

    Declare the Mutation model as:
        class PaymentMutation(core_models.UUIDModel, core_models.ObjectMutation):
        payment = models.ForeignKey(Payment, models.DO_NOTHING, related_name='mutations')
        mutation = models.ForeignKey(core_models.MutationLog, models.DO_NOTHING, related_name='payments')

        class Meta:
            managed = True
            db_table = "contribution_PaymentMutation"

    Call it like:
        client_mutation_id = data.get("client_mutation_id")
        payment = update_or_create_payment(data, user)
        PaymentMutation.object_mutated(user, client_mutation_id=client_mutation_id, payment=payment)
        return None

    Note that payment=payment, the name of the parameter gives the field name of the xxxMutation object to use
    and the value is the instance itself.
    """

    @classmethod
    def object_mutated(cls, user, mutation_log_id=None, client_mutation_id=None, *args, **kwargs):
        # This method should fail silently to not disrupt the actual mutation
        # noinspection PyBroadException
        try:
            args_models = {k + "_id": v.id for k, v in kwargs.items() if isinstance(v, models.Model)}
            if len(args_models) == 0 or len(args_models) > 1:
                logger.error("Trying to update ObjectMutationLink with several models in params: %s",
                             ", ".join(args_models.keys()))
                return
            if mutation_log_id:
                cls.objects.get_or_create(mutation_id=mutation_log_id, **args_models)
            elif client_mutation_id:
                mutations = MutationLog.objects \
                                .filter(client_mutation_id=client_mutation_id) \
                                .filter(user=user) \
                                .values_list("id", flat=True) \
                                .order_by("-request_date_time")[:2]  # Only ask for 2 for the warning, we'll only use 1
                if len(mutations) == 2:
                    # Warning because if done too often, this would cause performance issues in this query
                    logger.warning("Two or more mutations found for id %s, using the most recent one",
                                   client_mutation_id)
                if len(mutations) == 0:
                    logger.debug("No mutation found for client_mutation_id %s, ignoring", client_mutation_id)
                    return
                cls.objects.get_or_create(mutation_id=mutations[0], **args_models)
            else:
                logger.warning(
                    "Trying to update a %s without either mutation id or client_mutation_id, ignoring", cls.__name__)
        except Exception as exc:
            # The mutation shouldn't fail because we couldn't store the UUID
            logger.error("Error updating the %s object", cls.__name__, exc_info=True)


class HistoryModelManager(models.Manager):
    """
        Custom manager that allows querying HistoryModel by uuid
    """

    def get_queryset(self):
        return super().get_queryset().annotate(uuid=F('id'))


class HistoryModel(DirtyFieldsMixin, models.Model):
    id = models.UUIDField(primary_key=True, db_column="UUID", default=None, editable=False)
    objects = HistoryModelManager()
    is_deleted = models.BooleanField(db_column="isDeleted", default=False)
    json_ext = FallbackJSONField(db_column="Json_ext", blank=True, null=True)
    date_created = DateTimeField(db_column="DateCreated", null=True)
    date_updated = DateTimeField(db_column="DateUpdated", null=True)
    user_created = models.ForeignKey(User, db_column="UserCreatedUUID", related_name='%(class)s_user_created',
                                     on_delete=models.deletion.DO_NOTHING, null=False)
    user_updated = models.ForeignKey(User, db_column="UserUpdatedUUID", related_name='%(class)s_user_updated',
                                     on_delete=models.deletion.DO_NOTHING, null=False)
    version = models.IntegerField(default=1)
    history = HistoricalRecords(
        inherit=True,
    )

    @property
    def uuid(self):
        return self.id

    @uuid.setter
    def uuid(self, v):
        self.id = v

    def set_pk(self):
        self.pk = uuid.uuid4()

    def save_history(self):
        pass

    def save(self, *args, **kwargs):
        # get the user data so as to assign later his uuid id in fields user_updated etc
        user = User.objects.get(**kwargs)
        from core import datetime
        now = datetime.datetime.now()
        # check if object has been newly created
        if self.id is None:
            # save the new object
            self.set_pk()
            self.user_created = user
            self.user_updated = user
            self.date_created = now
            self.date_updated = now
            return super(HistoryModel, self).save()
        if self.is_dirty(check_relationship=True):
            self.date_updated = now
            self.user_updated = user
            self.version = self.version + 1
            # check if we have business model
            if hasattr(self, "replacement_uuid"):
                if self.replacement_uuid is not None and 'replacement_uuid' not in self.get_dirty_fields():
                    raise ValidationError('Update error! You cannot update replaced entity')
            return super(HistoryModel, self).save()
        else:
            raise ValidationError('Record has not be updated - there are no changes in fields')

    def delete_history(self):
        pass

    def delete(self, *args, **kwargs):
        user = User.objects.get(**kwargs)
        if not self.is_dirty(check_relationship=True) and not self.is_deleted:
            from core import datetime
            now = datetime.datetime.now()
            self.date_updated = now
            self.user_updated = user
            self.version = self.version + 1
            self.is_deleted = True
            # check if we have business model
            if hasattr(self, "replacement_uuid"):
                # When a replacement entity is deleted, the link should be removed
                # from replaced entity so a new replacement could be generated
                replaced_entity = self.__class__.objects.filter(replacement_uuid=self.id).first()
                if replaced_entity:
                    replaced_entity.replacement_uuid = None
                    replaced_entity.save(username="admin")
            return super(HistoryModel, self).save()
        else:
            raise ValidationError(
                'Record has not be deactivating, the object is different and must be updated before deactivating')

    @classmethod
    def filter_queryset(cls, queryset=None):
        if queryset is None:
            queryset = cls.objects.all()
        queryset = queryset.filter()
        return queryset

    class Meta:
        abstract = True


class HistoryBusinessModel(HistoryModel):
    date_valid_from = DateTimeField(db_column="DateValidFrom", default=py_datetime.now)
    date_valid_to = DateTimeField(db_column="DateValidTo", blank=True, null=True)
    replacement_uuid = models.UUIDField(db_column="ReplacementUUID", null=True)

    def replace_object(self, data, **kwargs):
        # check if object was created and saved in database (having date_created field)
        if self.id is None:
            return None
        user = User.objects.get(**kwargs)
        # 1 step - create new entity
        new_entity = self._create_new_entity(user=user, data=data)
        # 2 step - update the fields for the entity to be replaced
        self._update_replaced_entity(user=user, uuid_from_new_entity=new_entity.id,
                                     date_valid_from_new_entity=new_entity.date_valid_from)

    def _create_new_entity(self, user, data):
        """1 step - create new entity"""
        from core import datetime
        now = datetime.datetime.now()
        new_entity = copy(self)
        new_entity.id = None
        new_entity.version = 1
        new_entity.date_valid_from = now
        new_entity.date_valid_to = None
        new_entity.replacement_uuid = None
        # replace the fiedls if there are any to update in new entity
        if "uuid" in data:
            data.pop('uuid')
        if len(data) > 0:
            [setattr(new_entity, key, data[key]) for key in data]
        if self.date_valid_from is None:
            raise ValidationError('Field date_valid_from should not be empty')
        new_entity.save(username=user.username)
        return new_entity

    def _update_replaced_entity(self, user, uuid_from_new_entity, date_valid_from_new_entity):
        """2 step - update the fields for the entity to be replaced"""
        # convert to datetime if the date_valid_from from new entity is date
        from core import datetime
        if not isinstance(date_valid_from_new_entity, datetime.datetime):
            date_valid_from_new_entity = datetime.datetime.combine(
                date_valid_from_new_entity,
                datetime.datetime.min.time()
            )
        if not self.is_dirty(check_relationship=True):
            if self.date_valid_to is not None:
                if date_valid_from_new_entity < self.date_valid_to:
                    self.date_valid_to = date_valid_from_new_entity
            else:
                self.date_valid_to = date_valid_from_new_entity
            self.replacement_uuid = uuid_from_new_entity
            self.save(username=user.username)
            return self
        else:
            raise ValidationError("Object is changed - it must be updated before being replaced")

    class Meta:
        abstract = True


class RoleMutation(UUIDModel, ObjectMutation):
    role = models.ForeignKey(Role, models.DO_NOTHING,
                             related_name='mutations')
    mutation = models.ForeignKey(
        MutationLog, models.DO_NOTHING, related_name='roles')

    class Meta:
        managed = True
        db_table = "core_RoleMutation"


class UserMutation(UUIDModel, ObjectMutation):
    core_user = models.ForeignKey(User, models.CASCADE, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='users')

    class Meta:
        managed = True
        db_table = "core_UserMutation"



import sys
import importlib
from django.test import TestCase
from .shared import datetimedelta


class SharedDatetimedeltaTest(TestCase):

    def test_abs(self):
        delta = datetimedelta(days=-7)
        self.assertEquals(abs(delta), datetimedelta(days=7))

        delta = datetimedelta(months=-7, days=-2)
        self.assertEquals(abs(delta), datetimedelta(months=7, days=2))

        delta = datetimedelta(years=-7, days=-3)
        self.assertEquals(abs(delta), datetimedelta(years=7, days=3))

        delta = datetimedelta(years=7, months=5, days=-3)
        self.assertEquals(abs(delta), datetimedelta(
            years=7, months=5, days=-3))

    def test_add_sub(self):
        d1 = datetimedelta(days=7)
        d2 = datetimedelta(days=5)
        self.assertEquals(datetimedelta(days=12), d1 + d2)
        self.assertEquals(datetimedelta(days=2), d1 - d2)

        self.assertEquals(datetimedelta(days=7), +d1)
        self.assertEquals(datetimedelta(days=-7), -d1)

    def test_le_lt_ge_gt(self):
        self.assertTrue(datetimedelta(days=1) < datetimedelta(days=2))
        self.assertTrue(datetimedelta(days=1) <= datetimedelta(days=2))
        self.assertTrue(datetimedelta(days=2) > datetimedelta(days=1))
        self.assertTrue(datetimedelta(days=2) >= datetimedelta(days=1))

    def test_hash(self):
        self.assertNotEquals(hash(datetimedelta(days=2)), -1)

    def test_bool(self):
        self.assertTrue(datetimedelta(days=2))
        self.assertTrue(datetimedelta(weeks=2))


# Generated by Django 3.0.14 on 2021-06-01 12:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0012_users_officers_admins'),
    ]

    operations = [
        migrations.RenameField(
            model_name='usermutation',
            old_name='user',
            new_name='core_user',
        ),
    ]



# Generated by Django 3.0.3 on 2021-03-24 15:28
import django
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0010_rolemutation'),
    ]

    operations = [
        migrations.DeleteModel(
            name='RoleMutation'
        ),
        migrations.CreateModel(
            name='RoleMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='roles',
                                               to='core.MutationLog')),
                ('role', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations',
                                           to='core.Role')),
            ],
            options={
                'db_table': 'core_RoleMutation',
                'managed': True,
            },
        ),
    ]



import core
import graphene
from django.db.models import Q
from django.utils.translation import gettext as _

__all__ = [
    "TimeUtils",
    "full_class_name",
    "comparable",
    "filter_validity",
    "prefix_filterset",
    "assert_string_length",
    "PATIENT_CATEGORY_MASK_MALE",
    "PATIENT_CATEGORY_MASK_FEMALE",
    "PATIENT_CATEGORY_MASK_ADULT",
    "PATIENT_CATEGORY_MASK_MINOR",
    "patient_category_mask",
    "ExtendedConnection",
    "get_scheduler_method_ref",
    "ExtendedRelayConnection",
]


class TimeUtils(object):

    @classmethod
    def now(cls):
        return core.datetime.datetime.now()

    @classmethod
    def date(cls):
        return core.datetime.datetime.date(cls.now())


def full_class_name(o):
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    return module + '.' + o.__class__.__name__


def comparable(cls):
    """ Class decorator providing generic comparison functionality """

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    cls.__eq__ = __eq__
    cls.__ne__ = __ne__
    return cls


def filter_validity(arg="validity", **kwargs):
    validity = kwargs.get(arg)
    if validity is None:
        return (
            Q(validity_from__lte=core.datetime.datetime.now()),
            Q(legacy_id__isnull=True),
        )
    return (
       Q(validity_from__lte=validity),
        Q(validity_to__isnull=True) | Q(validity_to__gte=validity),
    )


def filter_validity_business_model(arg='dateValidFrom__Gte', arg2='dateValidTo__Lte', **kwargs):
    date_valid_from = kwargs.get(arg)
    date_valid_to = kwargs.get(arg2)
    #default scenario
    if not date_valid_from and not date_valid_to:
        today = core.datetime.datetime.now()
        return __place_the_filters(date_start=today, date_end=None)

    # scenario - only date valid to set
    if not date_valid_from and date_valid_to:
        today = core.datetime.datetime.now()
        oldest = min([today, date_valid_to])
        return __place_the_filters(date_start=oldest, date_end=date_valid_to)

    # scenario - only date valid from
    if date_valid_from and not date_valid_to:
        return __place_the_filters(date_start=date_valid_from, date_end=None)

    # scenario - both filters set
    if date_valid_from and date_valid_to:
        return __place_the_filters(date_start=date_valid_from, date_end=date_valid_to)


def __place_the_filters(date_start, date_end):
    """funtion related to 'filter_validity_business_model'
    function so as to set up the chosen filters
    to filter the validity of the entity
    """
    if not date_end:
        return (
            Q(date_valid_from__isnull=False),
            Q(date_valid_to__isnull=True) | Q(date_valid_to__gte=date_start)
        )
    return (
        Q(date_valid_from__lte=date_end),
        Q(date_valid_to__isnull=True) | Q(date_valid_to__gte=date_start)
    )


def append_validity_filter(**kwargs):
    default_filter = kwargs.get('applyDefaultValidityFilter', False)
    date_valid_from = kwargs.get('dateValidFrom__Gte', None)
    date_valid_to = kwargs.get('dateValidTo__Lte', None)
    filters = []
    # check if we can use default filter validity
    if date_valid_from is None and date_valid_to is None:
        if default_filter:
            filters = [*filter_validity_business_model(**kwargs)]
        else:
            filters = []
    else:
        filters = [*filter_validity_business_model(**kwargs)]
    return filters


def filter_is_deleted(arg='is_deleted', **kwargs):
    is_deleted = kwargs.get(arg)
    if is_deleted is None:
        is_deleted = False
    return (
        Q(is_deleted=is_deleted)
    )


def prefix_filterset(prefix, filterset):
    if type(filterset) is dict:
        return {(prefix + k): v for k, v in filterset.items()}
    elif type(filterset) is list:
        return [(prefix + x) for x in filterset]
    else:
        return filterset


def assert_string_length(str_value, length):
    if length and len(str_value) > length:
        raise Exception(
            _("core.string.over_max_length") % {
                'value': str_value, 'max_length': length})


PATIENT_CATEGORY_MASK_MALE = 1
PATIENT_CATEGORY_MASK_FEMALE = 2
PATIENT_CATEGORY_MASK_ADULT = 4
PATIENT_CATEGORY_MASK_MINOR = 8


def patient_category_mask(insuree, target_date):
    if type(target_date) is str:
        from core import datetime
        # TODO: this should be nicer
        target_date = datetime.date(*[int(x) for x in target_date.split("-")])
    mask = 0
    if not insuree.gender:
        raise NotImplementedError(_("core.insuree.unknown_gender"))
    if not insuree.dob:
        raise NotImplementedError(_("core.insuree.unknown_dob"))

    if insuree.gender.code in ('M', 'O'):
        mask = mask | PATIENT_CATEGORY_MASK_MALE
    else:
        mask = mask | PATIENT_CATEGORY_MASK_FEMALE

    if insuree.is_adult(target_date):
        mask = mask | PATIENT_CATEGORY_MASK_ADULT
    else:
        mask = mask | PATIENT_CATEGORY_MASK_MINOR
    return mask


class ExtendedConnection(graphene.Connection):
    """
    Adds total_count and edge_count to Graphene connections. To use, simply add to the
    Graphene object definition Meta:
    `connection_class = ExtendedConnection`
    """
    class Meta:
        abstract = True

    total_count = graphene.Int()
    edge_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs):
        return self.length

    def resolve_edge_count(self, info, **kwargs):
        return len(self.edges)


def get_scheduler_method_ref(name):
    """
    Use to retrieve the method reference from a str name. This is necessary when the module cannot be imported from
    that location.
    :param name: claim.module.submodule.method or similar name
    :return: reference to the method
    """
    split_name = name.split(".")
    module = __import__(".".join(split_name[:-1]))
    for subitem in split_name[1:]:
        module = getattr(module, subitem)
    return module


class ExtendedRelayConnection(graphene.relay.Connection):
    """
    Adds total_count and edge_count to Graphene Relay connections.
    """
    class Meta:
        abstract = True

    total_count = graphene.Int()
    edge_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs):
        return len(self.iterable)

    def resolve_edge_count(self, info, **kwargs):
        return len(self.edges)


def get_first_or_default_language():
    from core.models import Language
    sorted_languages = Language.objects.filter(sort_order__isnull=False)
    if sorted_languages.exists():
        return sorted_languages.order_by('sort_order').first()
    else:
        return Language.objects.first()



# Generated by Django 2.1.8 on 2019-04-15 14:35

import core.fields
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0009_alter_user_last_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='InteractiveUser',
            fields=[
                ('id', models.AutoField(db_column='UserID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('last_name', models.CharField(db_column='LastName', max_length=100)),
                ('other_names', models.CharField(db_column='OtherNames', max_length=100)),
                ('phone', models.CharField(blank=True, db_column='Phone', max_length=50, null=True)),
                ('login_name', models.CharField(db_column='LoginName', max_length=25)),
                ('role_id', models.IntegerField(db_column='RoleID')),
                ('health_facility_id', models.IntegerField(blank=True, db_column='HFID', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('email', models.CharField(blank=True, db_column='EmailId', max_length=200, null=True)),
            ],
            options={
                'db_table': 'tblUsers',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='Language',
            fields=[
                ('code', models.CharField(db_column='LanguageCode', max_length=2, primary_key=True, serialize=False)),
                ('name', models.CharField(db_column='LanguageName', max_length=50)),
                ('sort_order', models.IntegerField(blank=True, db_column='SortOrder', null=True)),
            ],
            options={
                'db_table': 'tblLanguages',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('username', models.CharField(max_length=25, unique=True)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('i_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.InteractiveUser')),
            ],
            options={
                'db_table': 'core_User',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='ModuleConfiguration',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('module', models.CharField(max_length=20)),
                ('version', models.CharField(max_length=10)),
                ('config', models.TextField()),
            ],
            options={
                'db_table': 'core_ModuleConfiguration',
            },
        ),
        migrations.CreateModel(
            name='TechnicalUser',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('username', models.CharField(max_length=150, unique=True)),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
                ('is_staff', models.BooleanField(default=False)),
                ('is_superuser', models.BooleanField(default=False)),
                ('validity_from', models.DateTimeField(blank=True, null=True)),
                ('validity_to', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'db_table': 'core_TechnicalUser',
                'managed': True,
            },
        ),
        migrations.AddField(
            model_name='user',
            name='t_user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.TechnicalUser'),
        ),
        migrations.AddField(
            model_name='user',
            name='user_permissions',
            field=models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions'),
        ),
    ]



import abc


class AbsCalculationRule(object,  metaclass=abc.ABCMeta):

    @classmethod
    def get_version(cls):
        return type(cls)._version

    @classmethod
    def set_version(cls, val):
        type(cls)._version = val

    version = property(get_version, set_version)

    @classmethod
    def get_uuid(cls):
        return type(cls)._uuid

    @classmethod
    def set_uuid(cls, val):
        type(cls)._uuid = val

    uuid = property(get_uuid, set_uuid)

    @classmethod
    def get_calculation_rule_name(cls):
        return type(cls)._calculation_rule_name

    @classmethod
    def set_calculation_rule_name(cls, val):
        type(cls)._calculation_rule_name = val

    calculation_rule_name = property(get_calculation_rule_name, set_calculation_rule_name)

    @classmethod
    def get_description(cls):
        return type(cls)._description

    @classmethod
    def set_description(cls, val):
        type(cls)._description = val

    description = property(get_description, set_description)

    @classmethod
    def get_impacted_class_parameter(cls):
        return type(cls)._impacted_class_parameter

    @classmethod
    def set_impacted_class_parameter(cls, val):
        type(cls)._impacted_class_parameter = val

    impacted_class_parameter = property(get_impacted_class_parameter, set_impacted_class_parameter)

    @classmethod
    @abc.abstractmethod
    def ready(cls):
        return

    @classmethod
    @abc.abstractmethod
    def check_calculation(cls, instance):
        return

    @classmethod
    @abc.abstractmethod
    def active_for_object(cls, instance, context):
        return

    @classmethod
    @abc.abstractmethod
    def calculate(cls, instance, *args):
        return

    @classmethod
    @abc.abstractmethod
    def get_linked_class(cls, sender, class_name, **kwargs):
        return

    @classmethod
    def get_rule_name(cls, sender, class_name, **kwargs):
        for object_class in cls.impacted_class_parameter:
            if object_class["class"] == class_name:
                return cls

    @classmethod
    def get_rule_details(cls, sender, class_name, **kwargs):
        for object_class in cls.impacted_class_parameter:
            if object_class["class"] == class_name:
                return object_class

    @classmethod
    def get_parameters(cls, sender, class_name, instance, **kwargs):
        """
            class_name is the class name of the object where the calculation param need to be added
            instance is where the link with a calculation need to be found,
            like the CPB in case of PH insuree or Contract Details
            return a list only with rule details that matches step 1 and 2
        """
        rule_details = cls.get_rule_details(sender=sender, class_name=class_name)
        if rule_details:
            if cls.check_calculation(instance=instance):
                return rule_details["parameters"] if "parameters" in rule_details else []

    @classmethod
    def run_calculation_rules(cls, sender, instance, user, context, **kwargs):
        """
            this function will send a signal and the rules will
            reply if they have object matching the classname in their list of object
        """
        list_class = cls.get_linked_class(sender=sender, class_name=instance.__class__.__name__)
        if len(list_class) > 0:
            rule_details = cls.get_rule_details(class_name=list_class[0], sender=sender)
            if rule_details:
                if cls.active_for_object(instance=instance, context=context):
                    result = cls.calculate(instance=instance)
                    return result



import sys
import importlib
import core
from django.test import TestCase
from datetime import date as py_date
from datetime import datetime as py_datetime
from .shared import is_midnight, datetimedelta


class SharedUtilsTest(TestCase):
    def setUp(self):
        core.calendar = importlib.import_module(
            '.calendars.ad_calendar', 'core')
        core.datetime = importlib.import_module(
            '.datetimes.ad_datetime', 'core')

    def test_is_midnight(self):
        self.assertTrue(is_midnight(
            core.datetime.datetime(2019, 4, 29, 0, 0, 0, 0)))
        self.assertTrue(is_midnight(core.datetime.date(2019, 4, 29)))
        self.assertFalse(is_midnight(
            core.datetime.datetime(2019, 4, 29, 11, 23, 56, 987654)))
        self.assertFalse(is_midnight(
            core.datetime.datetime(2019, 4, 29, 0, 23, 56, 987654)))
        self.assertFalse(is_midnight(
            core.datetime.datetime(2019, 4, 29, 0, 0, 56, 987654)))
        self.assertFalse(is_midnight(
            core.datetime.datetime(2019, 4, 29, 0, 0, 0, 987654)))

class AdDateTestCase(TestCase):
    def setUp(self):
        core.calendar = importlib.import_module(
            '.calendars.ad_calendar', 'core')
        core.datetime = importlib.import_module(
            '.datetimes.ad_datetime', 'core')

    def test_ad_date(self):
        py_dt = py_date(2020, 1, 13)
        ad_dt = core.datetime.date(2020, 1, 13)
        self.assertEqual(py_dt, ad_dt)

    def test_repr(self):
        ad_dt = core.datetime.date(2020, 1, 13)
        self.assertEqual(
            repr(ad_dt), 'core.datetimes.ad_datetime.date(2020, 1, 13)')

    def test_str(self):
        ad_dt = core.datetime.date(2020, 1, 13)
        self.assertEqual(str(ad_dt), '2020-01-13')

    def test_from_ad_datetime(self):
        db_dt = core.datetime.date.from_ad_datetime(py_datetime(2020, 1, 13))
        ad_dt = core.datetime.date(2020, 1, 13)
        self.assertEqual(db_dt, ad_dt)

        self.assertIsNone(core.datetime.date.from_ad_datetime(None))

    def test_to_ad_datetime(self):
        ad_dt = core.datetime.date(2020, 1, 13)
        db_dt = ad_dt.to_ad_datetime()
        self.assertEqual(db_dt, py_datetime(2020, 1, 13))

    def test_raw_iso_format(self):
        dt = core.datetime.date(2019, 3, 22)
        self.assertEquals("2019-03-22", dt.raw_isoformat())

    def test_ad_iso_format(self):
        dt = core.datetime.date(2019, 3, 22)
        self.assertEquals("2019-03-22", dt.ad_isoformat())

    def test_add_sub_years(self):
        dt = core.datetime.date(2020, 10, 22)
        self.assertEquals(core.datetime.date(2022, 10, 22),
                          (dt + datetimedelta(years=2)))
        self.assertEquals(core.datetime.date(2018, 10, 22),
                          (dt + datetimedelta(years=-2)))
        self.assertEquals(core.datetime.date(2022, 10, 22),
                          (dt - datetimedelta(years=-2)))
        self.assertEquals(core.datetime.date(2018, 10, 22),
                          (dt - datetimedelta(years=2)))
        dt = core.datetime.date(2020, 10, 29)
        self.assertEquals(core.datetime.date(2021, 10, 29),
                          (dt + datetimedelta(years=1)))
        dt = core.datetime.date(2024, 2, 29)
        self.assertEquals(core.datetime.date(2025, 2, 28),
                          (dt + datetimedelta(years=1)))

    def test_add_sub_months(self):
        dt = core.datetime.date(2020, 10, 22)
        self.assertEquals(core.datetime.date(2020, 12, 22),
                          (dt + datetimedelta(months=2)))
        self.assertEquals(core.datetime.date(2020, 8, 22),
                          (dt + datetimedelta(months=-2)))
        self.assertEquals(core.datetime.date(2020, 8, 22),
                          (dt - datetimedelta(months=2)))
        self.assertEquals(core.datetime.date(2020, 12, 22),
                          (dt - datetimedelta(months=-2)))

    def test_add_sub_days(self):
        dt = core.datetime.date(2019, 3, 22)
        self.assertEquals(core.datetime.date(2019, 4, 5),
                          (dt + datetimedelta(days=14)))
        self.assertEquals(core.datetime.date(2019, 4, 5),
                          (dt - datetimedelta(days=-14)))

        dt = core.datetime.date(2019, 3, 11)
        self.assertEquals(core.datetime.date(2019, 2, 25),
                          (dt - datetimedelta(days=14)))
        self.assertEquals(core.datetime.date(2019, 2, 25),
                          (dt + datetimedelta(days=-14)))

    def test_add_sub_mixed(self):
        dt = core.datetime.date(2020, 10, 22)
        self.assertEquals(
            core.datetime.date(2020, 12, 6),
            (dt + datetimedelta(months=1, days=14)))
        self.assertEquals(
            core.datetime.date(2021, 2, 28),
            (dt + 4 * datetimedelta(months=1, days=1, hours=14, seconds=36)))
        self.assertEquals(
            core.datetime.datetime(2020, 10, 22),
            (dt + datetimedelta(days=-1, hours=24)))
        self.assertEquals(
            core.datetime.date(2021, 2, 20),
            (dt + 4 * datetimedelta(months=1, days=-1, hours=14, seconds=36)))

    def test_eq_gt_ge_lt_le(self):
        dt = core.datetime.date(2019, 3, 11)
        dt_bis = core.datetime.date(2019, 3, 11)
        dt_after = core.datetime.date(2019, 3, 12)
        self.assertTrue(dt == dt_bis)
        self.assertTrue(dt >= dt_bis)
        self.assertTrue(dt <= dt_bis)
        self.assertTrue(dt < dt_after)
        self.assertTrue(dt <= dt_after)
        self.assertTrue(dt_after > dt)
        self.assertTrue(dt_after >= dt)

    def test_diff(self):
        ad_dt_1 = core.datetime.date(2019, 12, 7)
        ad_dt_2 = core.datetime.date(2019, 12, 8)
        self.assertEqual(ad_dt_2 - ad_dt_1, datetimedelta(days=1))

        ad_dt_1 = core.datetime.date(2019, 12, 7)
        ad_dt_2 = core.datetime.date(2020, 1, 7)
        self.assertEqual(ad_dt_2 - ad_dt_1, datetimedelta(days=31))


class AdDatetimeTestCase(TestCase):
    def setUp(self):
        core.calendar = importlib.import_module(
            '.calendars.ad_calendar', 'core')
        core.datetime = importlib.import_module(
            '.datetimes.ad_datetime', 'core')

    def test_ad_datetime(self):
        py_dt = py_datetime(2020, 1, 13, 10, 9, 55, 728267)
        ad_dt = core.datetime.datetime(2020, 1, 13, 10, 9, 55, 728267)
        self.assertEqual(py_dt, ad_dt)

    def test_repr(self):
        ad_dt = core.datetime.datetime(2020, 1, 13, 10, 9, 55, 728267)
        self.assertEqual(repr(
            ad_dt), 'core.datetimes.ad_datetime.datetime(2020, 1, 13, 10, 9, 55, 728267)')

    def test_str(self):
        ad_dt = core.datetime.datetime(2020, 1, 13, 10, 9, 55, 728267)
        self.assertEqual(str(ad_dt), '2020-01-13 10:09:55.728267')

    def test_add_sub_years(self):
        dt = core.datetime.datetime(2020, 10, 22, 14, 25, 44, 876134)
        self.assertEquals(
            core.datetime.datetime(2022, 10, 22, 14, 25, 44, 876134),
            (dt + datetimedelta(years=2)))
        self.assertEquals(
            core.datetime.datetime(2018, 10, 22, 14, 25, 44, 876134),
            (dt + datetimedelta(years=-2)))
        self.assertEquals(
            core.datetime.datetime(2022, 10, 22, 14, 25, 44, 876134),
            (dt - datetimedelta(years=-2)))
        self.assertEquals(
            core.datetime.datetime(2018, 10, 22, 14, 25, 44, 876134),
            (dt - datetimedelta(years=2)))

    def test_add_sub_months(self):
        dt = core.datetime.datetime(2020, 10, 22, 14, 25, 44, 876134)
        self.assertEquals(
            core.datetime.datetime(2020, 12, 22, 14, 25, 44, 876134),
            (dt + datetimedelta(months=2)))
        self.assertEquals(
            core.datetime.datetime(2020, 8, 22, 14, 25, 44, 876134),
            (dt + datetimedelta(months=-2)))
        self.assertEquals(
            core.datetime.datetime(2020, 8, 22, 14, 25, 44, 876134),
            (dt - datetimedelta(months=2)))
        self.assertEquals(
            core.datetime.datetime(2020, 12, 22, 14, 25, 44, 876134),
            (dt - datetimedelta(months=-2)))

    def test_add_sub_days(self):
        dt = core.datetime.datetime(2019, 3, 22)
        self.assertEquals(core.datetime.datetime(2019, 4, 5),
                          (dt + datetimedelta(days=14)))
        self.assertEquals(core.datetime.datetime(2019, 4, 5),
                          (dt - datetimedelta(days=-14)))

        dt = core.datetime.datetime(2019, 3, 11)
        self.assertEquals(core.datetime.datetime(2019, 2, 25),
                          (dt - datetimedelta(days=14)))
        self.assertEquals(core.datetime.datetime(2019, 2, 25),
                          (dt + datetimedelta(days=-14)))

    def test_add_sub_mixed(self):
        dt = core.datetime.datetime(2019, 10, 22, 5, 33, 24, 786222)
        self.assertEquals(
            core.datetime.datetime(2019, 12, 6, 5, 33, 24, 786222),
            (dt + datetimedelta(months=1, days=14)))
        self.assertEquals(
            core.datetime.datetime(2020, 2, 28, 13, 35, 48, 786222),
            (dt + 4 * datetimedelta(months=1, days=1, hours=14, seconds=36)))
        self.assertEquals(
            core.datetime.datetime(2019, 10, 22, 5, 33, 24, 786222),
            (dt + datetimedelta(days=-1, hours=24)))
        self.assertEquals(
            core.datetime.datetime(2020, 2, 20, 13, 35, 48, 786222),
            (dt + 4 * datetimedelta(months=1, days=-1, hours=14, seconds=36)))

    def test_eq_gt_ge_lt_le(self):
        dt = core.datetime.datetime(2019, 3, 11)
        dt_bis = core.datetime.datetime(2019, 3, 11)
        dt_after = core.datetime.datetime(2019, 3, 12)
        self.assertTrue(dt == dt_bis)
        self.assertTrue(dt >= dt_bis)
        self.assertTrue(dt <= dt_bis)
        self.assertTrue(dt < dt_after)
        self.assertTrue(dt <= dt_after)
        self.assertTrue(dt_after > dt)
        self.assertTrue(dt_after >= dt)

    def test_diff(self):
        ne_dt_1 = core.datetime.datetime(2019, 12, 7)
        ne_dt_2 = core.datetime.datetime(2019, 12, 8)
        self.assertEqual(ne_dt_2 - ne_dt_1, datetimedelta(days=1))

        ne_dt_1 = core.datetime.datetime(2019, 12, 7, 11, 5, 34, 999999)
        ne_dt_2 = core.datetime.datetime(2020, 1, 7, 11, 7, 34, 999999)
        self.assertEqual(ne_dt_2 - ne_dt_1,
                         datetimedelta(days=31, minutes=2))




class ObjectNotExistException(Exception):

    def __init__(self, model, uuid):
        self.model = model
        self.uuid = uuid
        self.message = '{model} with uuid {uuid} does not exist'\
            .format(model=self.model, uuid=self.uuid)
        super().__init__(self.message)


class MutationValidationException(Exception):

    def __init__(self, mutated_object, *invalidation_reasons):
        self.mutated_object = mutated_object
        self.errors = invalidation_reasons
        self.message = self._build_error_msg()
        super().__init__(self.message)

    def _build_error_msg(self):
        msg = ",\n".join(self.errors)
        return "Validation during a mutation has failed, reason:\n" \
               "{msg}".format(msg=msg)



# Generated by Django 3.0.3 on 2021-05-26 21:59

import core.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0012_item_service_jsonExtField'),
        ('core', '0011_auto_20210324_1528'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='claim_admin',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='claim.ClaimAdmin'),
        ),
        migrations.AddField(
            model_name='user',
            name='officer',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Officer'),
        ),
        migrations.CreateModel(
            name='UserMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='users', to='core.MutationLog')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'core_UserMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.RunSQL(sql="""
            insert into core_User (id, username, i_user_id, t_user_id, officer_id, claim_admin_id)
            select replace(lower(newid()), '-', ''), o.code, u.UserID, null, max(o.OfficerID), null
            from tblOfficer o
            left join tblUsers u on o.code = u.LoginName
            where not exists (select 1 from core_User where username=o.code)
            and o.ValidityTo is null and u.ValidityTo is null
            group by o.code, u.UserID;
            update core_User
            set officer_id=maxofficer.maxid
            from core_User cu inner join
                (select code, max(OfficerID) as maxid from tblOfficer where validityTo is null group by code) maxofficer
                    on code=cu.username
            where cu.officer_id is null;
            insert into core_User (id, username, i_user_id, t_user_id, officer_id, claim_admin_id)
            select replace(lower(newid()), '-', '') as uuid, ca.ClaimAdminCode, u.UserID, null as t, null as o, max(ca.ClaimAdminId) as max_id
            from tblClaimAdmin ca
            left join tblUsers u on ca.ClaimAdminCode = u.LoginName
            where not exists (select 1 from core_User where username=ca.ClaimAdminCode)
            and ca.ValidityTo is null and u.ValidityTo is null
            group by ca.ClaimAdminCode, u.UserID;
            update core_User
            set claim_admin_id=maxca.maxid
            from core_User cu inner join
                (select ClaimAdminCode, max(ClaimAdminId) as maxid from tblClaimAdmin where validityTo is null group by ClaimAdminCode) maxca
                    on ClaimAdminCode=cu.username
            where cu.claim_admin_id is null;
        """, reverse_sql="")
    ]



from django.db import transaction

from core import TimeUtils
from core.schema import OpenIMISMutation
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from core.gql.gql_mutations import ObjectNotExistException


class BaseMutation(OpenIMISMutation):

    @property
    def _mutation_class(self):
        raise NotImplementedError()

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            cls._validate_mutation(user, **data)
            mutation_result = cls._mutate(user, **data)
            return mutation_result
        except Exception as exc:
            return [{
                'message': "Failed to process {} mutation".format(cls._mutation_class),
                'detail': str(exc)}]

    @classmethod
    def _validate_mutation(cls, user, **data):
        raise NotImplementedError()

    @classmethod
    def _mutate(cls, user, **data):
        raise NotImplementedError()


class BaseDeleteMutation(BaseMutation):

    class Input:
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        output = []
        if "uuids" in data:
           for uuid in data["uuids"]:
               deletion_result = super(BaseDeleteMutation, cls)\
                   .async_mutate(user, uuid=uuid)
               deletion_result = [None] if deletion_result is None else ["error"]
               output += deletion_result
           return None if output == [None] * len(output) else output


class BaseReplaceMutation(BaseMutation):

    class Input:
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            cls._validate_mutation(user, **data)
            mutation_result = cls._mutate(user, **data)
            return mutation_result
        except Exception as exc:
            return [{
                'message': "Failed to process {} mutation".format(cls._mutation_class),
                'detail': str(exc)}]


class BaseCreateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        data['user_created'] = user.id_for_audit
        data['date_created'] = TimeUtils.now()

        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        cls.create_object(data)

    @classmethod
    def create_object(cls, object_data):
        obj = cls._model.objects.create(**object_data)
        obj.save()
        return obj


class BaseUpdateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

        obj_uuid = data['uuid']
        if cls._model.objects.filter(uuid=data['uuid']).first() is None:
            cls._object_not_exist_exception(obj_uuid=obj_uuid)

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        data['user_updated'] = user.id_for_audit
        data['date_updated'] = TimeUtils.now()
        updated_object = cls._model.objects.filter(uuid=data['uuid']).first()
        [setattr(updated_object, key, data[key]) for key in data]
        cls.update_object(updated_object)

    @classmethod
    def update_object(cls, object_to_update):
        object_to_update.save_history()
        object_to_update.save()
        return object_to_update


class BaseDeleteMutationMixin:
    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        cls._validate_user(user)

    @classmethod
    def _validate_user(cls, user):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, uuid):
        object_to_delete = cls._model.objects.filter(uuid=uuid).first()

        if object_to_delete is None:
            cls._object_not_exist_exception(uuid)
        else:
            object_to_delete.delete_history()


class BaseHistoryModelCreateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        cls.create_object(user=user, object_data=data)

    @classmethod
    def create_object(cls, user, object_data):
        obj = cls._model(**object_data)
        obj.save(username=user.username)
        return obj


class BaseHistoryModelUpdateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")
        obj_uuid = data['id']
        if cls._model.objects.filter(id=data['id']).first() is None:
            cls._object_not_exist_exception(obj_uuid=obj_uuid)

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        updated_object = cls._model.objects.filter(id=data['id']).first()
        [setattr(updated_object, key, data[key]) for key in data]
        cls.update_object(user=user, object_to_update=updated_object)

    @classmethod
    def update_object(cls, user, object_to_update):
        object_to_update.save(username=user.username)
        return object_to_update


class BaseHistoryModelDeleteMutationMixin:
    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        cls._validate_user(user)

    @classmethod
    def _validate_user(cls, user):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        id_ = data.get('uuid', None) or data.get('id', None)
        ids = data.get('uuids', None) or data.get('ids', None)
        if id_:
            cls.__delete_single_obj(user=user, id_=id_)
        elif ids:
            with transaction.atomic():
                for id_ in ids:
                    cls.__delete_single_obj(user, id_)

    @classmethod
    def __delete_single_obj(cls, user, id_):
        object_to_delete = cls._model.objects.filter(id=id_).first()
        if object_to_delete is None:
            cls._object_not_exist_exception(id_)
        else:
            object_to_delete.delete(username=user.username)


class BaseHistoryModelReplaceMutationMixin:
    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        cls._validate_user(user)

    @classmethod
    def _validate_user(cls, user):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        object_to_replace = cls._model.objects.filter(id=data['uuid']).first()
        if object_to_replace is None:
            cls._object_not_exist_exception(data['uuid'])
        else:
            object_to_replace.replace_object(data=data, username=user.username)



import sys
import importlib
import core
from django.test import TestCase
from ..datetimes.ad_datetime import date
from .ad_calendar import *


class CalendarTestCase(TestCase):
    def setUp(self):
        core.calendar = importlib.import_module(
            '.calendars.ad_calendar', 'core')
        core.datetime = importlib.import_module(
            '.datetimes.ad_datetime', 'core')

    def test_weekfirstday(self):
        dte = core.datetime.date(2019, 3, 19)
        fwd = core.calendar.weekfirstday(dte)
        self.assertEqual(fwd, core.datetime.date(2019, 3, 18))

    def test_weeklastday(self):
        dte = core.datetime.date(2019, 3, 19)
        fwd = core.calendar.weeklastday(dte)
        self.assertEqual(fwd, core.datetime.date(2019, 3, 24))

    def test_monthfirstday(self):
        dte = core.calendar.monthfirstday(2019, 3)
        self.assertEqual(dte, core.datetime.date(2019, 3, 1))

    def test_monthlastday(self):
        dte = core.calendar.monthlastday(2019, 3)
        self.assertEqual(dte, core.datetime.date(2019, 3, 31))

    def test_yearfirstday(self):
        dte = core.calendar.yearfirstday(2019)
        self.assertEqual(dte, core.datetime.date(2019, 1, 1))

    def test_yearlastday(self):
        dte = core.calendar.yearlastday(2019)
        self.assertEqual(dte, core.datetime.date(2019, 12, 31))

    def test_monthdayscount(self):
        self.assertEqual(29, core.calendar.monthdayscount(2020, 2))
        self.assertEqual(31, core.calendar.monthdayscount(2019, 12))

    def test_yeardayscount(self):
        self.assertEqual(365, core.calendar.yeardayscount(2019))
        self.assertEqual(366, core.calendar.yeardayscount(2020))



import logging
from functools import lru_cache

from django.db import migrations
from core.models import RoleRight, Role

logger = logging.getLogger(__name__)

ROLE_RIGHTS_ID = [101201, 101101, 101001, 101105]
MANAGER_ROLE_IS_SYSTEM = 256  # By default missing rights are assigned to ClaimAdmin role


@lru_cache(maxsize=1)
def __get_role_owner() -> Role:
    return Role.objects.get(is_system=256, validity_to=None)


def __role_already_exists(right_id):
    sc = RoleRight.objects.filter(role__uuid=__get_role_owner().uuid, right_id=right_id)
    return sc.count() > 0


def create_role_right(apps, schema_editor):
    if schema_editor.connection.alias != 'default':
        return
    for right_id in ROLE_RIGHTS_ID:
        if __role_already_exists(right_id):
            logger.warning(F"Role right {right_id} already assigned for role {__get_role_owner().name}, skipping")
            return
        role_owner = Role.objects.get(is_system=256, validity_to=None)
        new_role = RoleRight(
            role=role_owner,
            right_id=right_id,
            audit_user_id=None,
        )
        new_role.save()

    # Your migration code goes here


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0014_users')
    ]

    operations = [
        migrations.RunPython(create_role_right),
    ]



import json
import logging
from functools import wraps

import django.db.models
from django.db.models import Q
from graphene_django import DjangoObjectType
from typing import Dict


def mutation_on_uuids_from_filter(django_object: django.db.models.Model,
                                  object_gql_type: DjangoObjectType,
                                  query_filters_field: str = 'additional_filters',
                                  explicit_filters_handlers: Dict[str, str] = None,
                                  return_objects: bool = False):
    """
    A decorator for async_mutate allowing use of filters instead of directly specifying the UUID in migrations.
    If data argument of async_mutate don't have 'uuids' key it tries to fetch objects by filters. As result uuids of
    filtered objects are added to data['uuids']. If 'return_objects' value is set to True,
    it adds filtered objects to data['filtered_objects'] instead of adding uuids.

    If in the incoming mutation the uuid's of the objects to be mutated are not specified directly, the decorator uses
    the value of the field of 'query_filters_field' field to build the query which is executed on 'django_object'
    and returns filtered objects uuids are used in mutation. Filters executed on queryset are built from the content
    of the 'object_gql_type' object's Metaclass filter_fields. If incoming filters are containing keys not included
    in the Metaclass, then this filters must be handled by explicit_filters_handlers.


    :param django_object: Django model to be filtered for uuids
    :param object_gql_type: DjangoObjectType with filter_fields in Metaclass,
    which is GQL equivalent to the django_object
    :param query_filters_field: A field from an incoming query that contains filter information. The value referenced
    by this field must be a JSON serialized to string.
    :param explicit_filters_handlers: Additional configured filters, if there are filters in the query that have no
    counterpart in object_gql_type they must be included here. Value of keys should be in django queryset notation.
    Example:
        Additional key 'services' with list of service codes can be included in explicit_filters_handlers
        as { 'services': 'services__service__code__in'...}
    :param return_objects: Optional argument if, set to True instead of adding uuid's do data['uuids'] objects are added
    to data['filtered_objects']
    :return: Queryset<[uuid]> containing with uuids of objects that were received after filtering
    """
    if explicit_filters_handlers is None:
        explicit_filters_handlers = {}

    available_filters = _build_filters_from_gql_filters(object_gql_type._meta.filter_fields)

    def inner_function(async_mutate):

        @wraps(async_mutate)
        def wrapper(cls, user, **data):
            if not data.get('uuids', None):
                args = json.loads(data[query_filters_field])

                q_filter = map_gql_to_django_filter(args, available_filters, explicit_filters_handlers)
                base_query = django_object \
                    .objects \
                    .filter(validity_to=None) \
                    .filter(q_filter) \

                if return_objects:
                    data['filtered_objects'] = base_query
                else:
                    uuids = base_query.values_list('uuid', flat=True).distinct()
                    data['uuids'] = uuids
            async_mutate(cls, user, **data)
        return wrapper
    return inner_function


def map_gql_to_django_filter(filters: dict, qgl_type_filters, explicit_handlers=None):
    if explicit_handlers is None:
        explicit_handlers = {}

    def __disable_notation(k):
        return k.lower().replace('_', '')

    mapped_filters = []
    for key, param in filters.items():
        try:
            if key in explicit_handlers.keys():
                mapped_filters.append(Q(**{explicit_handlers[key]: param}))
            else:
                filter_key = __disable_notation(key)
                django_filter = next((gql_key for gql_key in qgl_type_filters
                                      if __disable_notation(gql_key) == filter_key))
                mapped_filters.append(Q(**{django_filter: param}))
        except StopIteration as s:
            error_msg = f"Could not find mapping for filter key {key}, available keys are {qgl_type_filters}"
            logging.error(error_msg)
            raise ValueError(error_msg)

    query_statement = mapped_filters.pop()  # get first query object
    for next_filter in mapped_filters:
        query_statement &= next_filter  # join remaining filters

    return query_statement


def _build_filters_from_gql_filters(filter_fields):
    fields = []
    for field, compare_types in filter_fields.items():
        for compare_type in compare_types:
            if compare_type == 'exact':
                fields.append(field)
            else:
                query_filter = F"{field}__{compare_type.lower()}"
                fields.append(query_filter)
    return fields


def mutation_on_uuids_from_filter_business_model(django_object: django.db.models.Model,
                                                 object_gql_type: DjangoObjectType,
                                                 query_filters_field: str = 'extended_filters',
                                                 explicit_filters_handlers: Dict[str, str] = None,
                                                 return_objects: bool = False):
    """
    dedicated extended mutation from filter decorator dedicated for BusinessHistoryModel entities (used for example
    in Formal Sector entities). See doc string for mutation_on_uuids_from_filter to read more how this works.
    """

    if explicit_filters_handlers is None:
        explicit_filters_handlers = {}

    available_filters = _build_filters_from_gql_filters(object_gql_type._meta.filter_fields)

    def inner_function(async_mutate):
        @wraps(async_mutate)
        def wrapper(cls, user, **data):
            # check if uuids exists as a substring in one of the key
            key_values_match = [key for key, value in data.items() if 'uuids' in key]
            uuids_exists = False
            # check also if list of uuids contains any uuids if uuids key exists
            for key in key_values_match:
                if key in data:
                    if len(data[key]) > 0:
                        uuids_exists = True
            if not uuids_exists:
                args = json.loads(data[query_filters_field])
                #for contract entities
                # TODO: need to "pop" based on a section "advanced_search"
                amount_to = None
                amount_from = None
                if 'amountFrom' in args:
                    amount_from = args['amountFrom']
                    args.pop("amountFrom")
                if 'amountTo' in args:
                    amount_to = args['amountTo']
                    args.pop("amountTo")
                # remove validity filter if applied
                if 'dateValidFrom_Gte' in args:
                    args.pop("dateValidFrom_Gte")
                if 'dateValidTo_Lte' in args:
                    args.pop("dateValidTo_Lte")
                # remove isDeleted if applied
                if 'is_deleted' in args:
                    args.pop("is_deleted")

                q_filter = map_gql_to_django_filter(args, available_filters, explicit_filters_handlers)
                from core import datetime
                now = datetime.datetime.now()

                base_query = django_object \
                    .objects \
                    .filter(
                        Q(date_valid_from__lte=now),
                        Q(date_valid_to=None) | Q(date_valid_to__gte=now),
                        Q(is_deleted=False)
                    ) \
                    .filter(q_filter) \

                # if mutation is related to contract entities
                # TODO check type of amount filter to check if we can send signal
                # TODO: need to send signal for the "pop" based on a search section "advanced_search" (signal should have the advances_search section and the object name)
                #  (to be implemented in the future) to contract module
                if django_object.__name__ == 'Contract':
                    if amount_from or amount_to:
                        base_query = base_query.filter(_append_filter_amount(amount_from, amount_to))

                if return_objects:
                    data['filtered_objects'] = base_query
                else:
                    uuids = base_query.values_list('id', flat=True).distinct()
                    data['uuids'] = uuids
            async_mutate(cls, user, **data)
        return wrapper
    return inner_function


def _append_filter_amount(amount_from, amount_to):
    # TODO make a signal to contract module to not break modular approach and
    #  not repeat the same code from contract module

    status_notified = [1, 2]
    status_rectified = [4, 11, 3]
    status_due = [5, 6, 7, 8, 9, 10]

    # scenario - only amount_to set
    if not amount_from and amount_to:
        return (
             Q(amount_notified__lte=amount_to, state__in=status_notified) |
             Q(amount_rectified__lte=amount_to, state__in=status_rectified) |
             Q(amount_due__lte=amount_to, state__in=status_due)
        )

    # scenario - only amount_from set
    if amount_from and not amount_to:
        return (
            Q(amount_notified__gte=amount_from, state__in=status_notified) |
            Q(amount_rectified__gte=amount_from, state__in=status_rectified) |
            Q(amount_due__gte=amount_from, state__in=status_due)
        )

    # scenario - both filters set
    if amount_from and amount_to:
        return (
            Q(amount_notified__gte=amount_from, amount_notified__lte=amount_to, state__in=status_notified) |
            Q(amount_rectified__gte=amount_from, amount_rectified__lte=amount_to, state__in=status_rectified) |
            Q(amount_due__gte=amount_from, amount_due__lte=amount_to, state__in=status_due)
        )



from .jwt import jwt_decode_user_key

from django.apps import apps
from rest_framework.authentication import BaseAuthentication
from rest_framework import exceptions

import jwt
import logging

logger = logging.getLogger(__file__)


class JWTAuthentication(BaseAuthentication):
    """
        class to obtain token from header if it is provided
        and verify if this is correct/valid token
    """

    def authenticate(self, request):
        authorization_header = request.headers.get('Authorization')
        if not authorization_header:
            return None
        payload = None
        if authorization_header.lower().startswith('bearer '):
            bearer, access_token, *extra_words = authorization_header.split(' ')
            if len(extra_words) > 0:
                raise exceptions.AuthenticationFailed("Improper structure of token")
            try:
                payload = jwt_decode_user_key(token=access_token)
            except jwt.DecodeError:
                raise exceptions.AuthenticationFailed('Error on decoding token')
            except jwt.ExpiredSignatureError:
                raise exceptions.AuthenticationFailed('Access_token expired')
            except IndexError:
                raise exceptions.AuthenticationFailed('Token prefix missing')
            user = None
            if payload:
                user_class = apps.get_model("core", "User")
                user = user_class.objects \
                    .filter(username=payload.get("username")) \
                    .only("i_user__private_key") \
                    .first()
            if user is None:
                raise exceptions.AuthenticationFailed('User inactive or deleted/not existed.')
            if not user.is_active:
                raise exceptions.AuthenticationFailed('User is inactive')
        else:
            raise exceptions.AuthenticationFailed("Missing 'Bearer' prefix")

        self.enforce_csrf(request)

        return user, None

    def enforce_csrf(self, request):
        return  # To not perform the csrf during checking auth header




from core.models import Officer, InteractiveUser, User
from core.services import create_or_update_user_roles


def create_test_officer(valid=True, custom_props=None):
    officer = Officer.objects.create(
        **{
            "code": "TSTOFF",
            "last_name": "Officer",
            "other_names": "Test",
            "validity_to": None if valid else "2019-06-01",
            "audit_user_id": -1,
            **(custom_props if custom_props else {})
        }
    )
    return officer


def create_test_interactive_user(username, password="Test1234", roles=None, custom_props=None):
    if roles is None:
        roles = [7, 1, 2, 3, 4, 5, 6]
    i_user = InteractiveUser.objects.create(
        **{
            "language_id": "en",
            "last_name": "TestLastName",
            "other_names": "Test Other Names",
            "login_name": username,
            "audit_user_id": -1,
            "role_id": roles[0],
            **(custom_props if custom_props else {})
        }
    )
    i_user.set_password(password)
    i_user.save()
    create_or_update_user_roles(i_user, roles, None)
    return User.objects.create(
        username=username,
        i_user=i_user,
    )



import graphene
import location.gql_queries
from core import ExtendedConnection, filter_validity
from core.models import Officer, Role, RoleRight, UserRole, User, InteractiveUser, UserMutation, Language
from graphene_django import DjangoObjectType
from location.models import HealthFacility

from .utils import prefix_filterset


class OfficerGQLType(DjangoObjectType):
    """
    This type corresponds to the Enrolment Officer but is a bit more generic than just Enrolment.
    """
    class Meta:
        model = Officer
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "uuid": ["exact"],
            "code": ["exact", "icontains"],
            "last_name": ["exact", "icontains"],
            "other_names": ["exact", "icontains"],
            "dob": ["exact"],
        }
        connection_class = ExtendedConnection

    @classmethod
    def get_queryset(cls, queryset, info):
        queryset = queryset.filter(*filter_validity())
        return queryset


class RoleGQLType(DjangoObjectType):
    class Meta:
        model = Role
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "uuid": ["exact"],
            "name": ["exact", "istartswith", "icontains", "iexact"],
            "is_blocked": ["exact"],
        }
        connection_class = ExtendedConnection

    @classmethod
    def get_queryset(cls, queryset, info):
        return Role.get_queryset(queryset, info)


class UserRoleGQLType(DjangoObjectType):
    class Meta:
        model = UserRole
        filter_fields = {
            "id": ["exact"],
            "user": ["exact"],
            "role": ["exact"],
        }
        connection_class = ExtendedConnection

    @classmethod
    def get_queryset(cls, queryset, info):
        return Role.get_queryset(queryset, info)


class RoleRightGQLType(DjangoObjectType):
    class Meta:
        model = RoleRight
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "right_id": ["exact"],
            **prefix_filterset("role__", RoleGQLType._meta.filter_fields),
        }
        connection_class = ExtendedConnection

    @classmethod
    def get_queryset(cls, queryset, info):
        return RoleRight.get_queryset(queryset, info)


class InteractiveUserGQLType(DjangoObjectType):
    """
    The InteractiveUser represents the regular openIMIS allowed to connect to the web interface. Enrolment Officers
    and Claim Administrators can exist without having web access but when they do, they have a corresponding
    InteractiveUser, linked by their "code" aka "login_name"
    """
    language_id = graphene.String()
    health_facility = graphene.Field(
        location.gql_queries.HealthFacilityGQLType,
        description="Health Facility is not a foreign key in the database, this field resolves it manually, use only "
                    "if necessary.")
    roles = graphene.List(RoleGQLType, description="Same as userRoles but a straight list, without the M-N relation")

    class Meta:
        model = InteractiveUser
        interfaces = (graphene.relay.Node,)
        exclude = ("stored_password", "private_key")
        filter_fields = {
            "id": ["exact"],
            "uuid": ["exact"],
            "last_name": ["icontains"],
            "other_names": ["icontains"],
            "phone": ["iexact"],
            "login_name": ["iexact"],
            "email": ["iexact"],
            "is_associated": ["exact"],
            "language_id": ["exact"],
        }
        connection_class = ExtendedConnection

    def resolve_health_facility(self, info, **kwargs):
        if self.health_facility_id:
            return HealthFacility.get_queryset(None, info).filter(pk=self.health_facility_id).first()
        else:
            return None

    def resolve_roles(self, info, **kwargs):
        if self.user_roles:
            return Role.objects\
                .filter(validity_to__isnull=True)\
                .filter(user_roles__user_id=self.id, user_roles__validity_to__isnull=True)
        else:
            return None

    @classmethod
    def get_queryset(cls, queryset, info):
        return InteractiveUser.get_queryset(queryset, info)


class UserGQLType(DjangoObjectType):
    """
    This type provides an abstraction of the various user types, TechnicalUser and InteractiveUser. It corresponds
    to the core_User table added in the modular version. The TechnicalUser is for now not exposed here as it is not
    managed through this API.
    """
    client_mutation_id = graphene.String()

    class Meta:
        model = User
        filter_fields = {
            "id": ["exact"],
            "username": ["exact", "icontains"],
            **prefix_filterset("i_user__", InteractiveUserGQLType._meta.filter_fields),
            **prefix_filterset("officer__", OfficerGQLType._meta.filter_fields),
        }
        interfaces = (graphene.relay.Node,)
        connection_class = ExtendedConnection

    @classmethod
    def get_queryset(cls, queryset, info):
        return User.get_queryset(queryset, info)

    def resolve_client_mutation_id(self, info):
        user_mutation = self.mutations.select_related('mutation').filter(mutation__status=0).first()
        return user_mutation.mutation.client_mutation_id if user_mutation else None


class PermissionOpenImisGQLType(graphene.ObjectType):
    perms_name = graphene.String()
    perms_value = graphene.Int()


class ModulePermissionGQLType(graphene.ObjectType):
    module_name = graphene.String()
    permissions = graphene.List(PermissionOpenImisGQLType)


class ModulePermissionsListGQLType(graphene.ObjectType):
    module_perms_list = graphene.List(ModulePermissionGQLType)


class UserMutationGQLType(DjangoObjectType):
    """
    This intermediate object links Mutations to Users. Beware of the confusion between the user performing the mutation
    and the users affected by that mutation, the latter being listed in this object.
    """
    class Meta:
        model = UserMutation

        
class LanguageGQLType(DjangoObjectType):
    class Meta:
        model = Language
        filter_fields = {
            "language_code": ["exact"],
            "name": ["exact"],
            "sort_order": ["exact"]
        }

    @classmethod
    def get_queryset(cls, queryset, info):
        return queryset



import concurrent.futures
import asyncio
import logging
import time
from contextlib import asynccontextmanager
from functools import partial, wraps

from typing import Callable, Coroutine

from asgiref.sync import sync_to_async

from core.websocket.base_websocket_client import BaseWebSocketClient

logger = logging.getLogger(__name__)


class AsyncWebSocketClient(BaseWebSocketClient):
    """
    Asynchronous implementation of WebSocket client.
    """

    def add_action_on_receive(self, on_receive_func: Callable[[object], Coroutine]):
        self.recv_actions.append(on_receive_func)

    async def send(self, payload):
        super().send(payload)

    def _on_recv(self, message):
        # Through to build of websockets.WebSocketApp async cannot be used directly

        # tasks = asyncio.gather(**self.recv_actions)
        loop = asyncio.new_event_loop()
        try:
            for action in self.recv_actions:
                loop.run_until_complete(action(message))
        except Exception as e:
            logger.error(F'Failed to execute action, reason: {e}')
        finally:
            loop.close()
        if not self.recv_actions:
            self._default_action(message)

    @asynccontextmanager
    async def connect(self):
        """
        Allows to use asynchronous context manager.
        Example:
            async with client_instance.connect() as connection:
                connection.send("Hello world")

        """
        self.open_connection()
        try:
            yield self
        finally:
            self.close_connection()

    async def __aenter__(self):
        self.open_connection()
        await self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close_connection()



# Generated by Django 3.0.14 on 2021-06-15 16:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_users_api'),
    ]

    operations = [
        migrations.AlterField(
            model_name='usermutation',
            name='core_user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mutations', to=settings.AUTH_USER_MODEL),
        ),
    ]
